from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class WorkerLabelComparator(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    EQUAL: _ClassVar[WorkerLabelComparator]
    NOT_EQUAL: _ClassVar[WorkerLabelComparator]
    GREATER_THAN: _ClassVar[WorkerLabelComparator]
    GREATER_THAN_OR_EQUAL: _ClassVar[WorkerLabelComparator]
    LESS_THAN: _ClassVar[WorkerLabelComparator]
    LESS_THAN_OR_EQUAL: _ClassVar[WorkerLabelComparator]
EQUAL: WorkerLabelComparator
NOT_EQUAL: WorkerLabelComparator
GREATER_THAN: WorkerLabelComparator
GREATER_THAN_OR_EQUAL: WorkerLabelComparator
LESS_THAN: WorkerLabelComparator
LESS_THAN_OR_EQUAL: WorkerLabelComparator

class DesiredWorkerLabels(_message.Message):
    __slots__ = ("str_value", "int_value", "required", "comparator", "weight")
    STR_VALUE_FIELD_NUMBER: _ClassVar[int]
    INT_VALUE_FIELD_NUMBER: _ClassVar[int]
    REQUIRED_FIELD_NUMBER: _ClassVar[int]
    COMPARATOR_FIELD_NUMBER: _ClassVar[int]
    WEIGHT_FIELD_NUMBER: _ClassVar[int]
    str_value: str
    int_value: int
    required: bool
    comparator: WorkerLabelComparator
    weight: int
    def __init__(self, str_value: _Optional[str] = ..., int_value: _Optional[int] = ..., required: bool = ..., comparator: _Optional[_Union[WorkerLabelComparator, str]] = ..., weight: _Optional[int] = ...) -> None: ...

class TriggerWorkflowRequest(_message.Message):
    __slots__ = ("name", "input", "parent_id", "parent_task_run_external_id", "child_index", "child_key", "additional_metadata", "desired_worker_id", "priority", "desired_worker_labels")
    class DesiredWorkerLabelsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: DesiredWorkerLabels
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[DesiredWorkerLabels, _Mapping]] = ...) -> None: ...
    NAME_FIELD_NUMBER: _ClassVar[int]
    INPUT_FIELD_NUMBER: _ClassVar[int]
    PARENT_ID_FIELD_NUMBER: _ClassVar[int]
    PARENT_TASK_RUN_EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    CHILD_INDEX_FIELD_NUMBER: _ClassVar[int]
    CHILD_KEY_FIELD_NUMBER: _ClassVar[int]
    ADDITIONAL_METADATA_FIELD_NUMBER: _ClassVar[int]
    DESIRED_WORKER_ID_FIELD_NUMBER: _ClassVar[int]
    PRIORITY_FIELD_NUMBER: _ClassVar[int]
    DESIRED_WORKER_LABELS_FIELD_NUMBER: _ClassVar[int]
    name: str
    input: str
    parent_id: str
    parent_task_run_external_id: str
    child_index: int
    child_key: str
    additional_metadata: str
    desired_worker_id: str
    priority: int
    desired_worker_labels: _containers.MessageMap[str, DesiredWorkerLabels]
    def __init__(self, name: _Optional[str] = ..., input: _Optional[str] = ..., parent_id: _Optional[str] = ..., parent_task_run_external_id: _Optional[str] = ..., child_index: _Optional[int] = ..., child_key: _Optional[str] = ..., additional_metadata: _Optional[str] = ..., desired_worker_id: _Optional[str] = ..., priority: _Optional[int] = ..., desired_worker_labels: _Optional[_Mapping[str, DesiredWorkerLabels]] = ...) -> None: ...



"""Pre-eviction fallback implementations for DurableContext.

These methods support engines older than MIN_DURABLE_EVICTION_VERSION.
Remove this module when support for those engines is dropped.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from hatchet_sdk.clients.listeners.legacy.pre_eviction_durable_event_listener import (
    PreEvictionDurableEventListener,
    RegisterDurableEventRequest,
)
from hatchet_sdk.conditions import (
    OrGroup,
    SleepCondition,
    UserEventCondition,
    flatten_conditions,
)

if TYPE_CHECKING:
    from hatchet_sdk.context.context import DurableContext


async def aio_wait_for_pre_eviction(
    ctx: DurableContext,
    signal_key: str,
    *conditions: SleepCondition | UserEventCondition | OrGroup,
) -> dict[str, Any]:
    if not isinstance(ctx.durable_event_listener, PreEvictionDurableEventListener):
        raise TypeError(
            "Expected PreEvictionDurableEventListener, got "
            f"{type(ctx.durable_event_listener).__name__}"
        )

    task_id = ctx.step_run_id

    request = RegisterDurableEventRequest(
        task_id=task_id,
        signal_key=signal_key,
        conditions=flatten_conditions(list(conditions)),
        config=ctx.runs_client.client_config,
    )

    ctx.durable_event_listener.register_durable_event(request)

    return await ctx.durable_event_listener.result(task_id, signal_key)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: v1/shared/trigger.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'v1/shared/trigger.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17v1/shared/trigger.proto\x12\x02v1\"\xe8\x01\n\x13\x44\x65siredWorkerLabels\x12\x16\n\tstr_value\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x16\n\tint_value\x18\x02 \x01(\x05H\x01\x88\x01\x01\x12\x15\n\x08required\x18\x03 \x01(\x08H\x02\x88\x01\x01\x12\x32\n\ncomparator\x18\x04 \x01(\x0e\x32\x19.v1.WorkerLabelComparatorH\x03\x88\x01\x01\x12\x13\n\x06weight\x18\x05 \x01(\x05H\x04\x88\x01\x01\x42\x0c\n\n_str_valueB\x0c\n\n_int_valueB\x0b\n\t_requiredB\r\n\x0b_comparatorB\t\n\x07_weight\"\xb2\x04\n\x16TriggerWorkflowRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05input\x18\x02 \x01(\t\x12\x16\n\tparent_id\x18\x03 \x01(\tH\x00\x88\x01\x01\x12(\n\x1bparent_task_run_external_id\x18\x04 \x01(\tH\x01\x88\x01\x01\x12\x18\n\x0b\x63hild_index\x18\x05 \x01(\x05H\x02\x88\x01\x01\x12\x16\n\tchild_key\x18\x06 \x01(\tH\x03\x88\x01\x01\x12 \n\x13\x61\x64\x64itional_metadata\x18\x07 \x01(\tH\x04\x88\x01\x01\x12\x1e\n\x11\x64\x65sired_worker_id\x18\x08 \x01(\tH\x05\x88\x01\x01\x12\x15\n\x08priority\x18\t \x01(\x05H\x06\x88\x01\x01\x12R\n\x15\x64\x65sired_worker_labels\x18\n \x03(\x0b\x32\x33.v1.TriggerWorkflowRequest.DesiredWorkerLabelsEntry\x1aS\n\x18\x44\x65siredWorkerLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12&\n\x05value\x18\x02 \x01(\x0b\x32\x17.v1.DesiredWorkerLabels:\x02\x38\x01\x42\x0c\n\n_parent_idB\x1e\n\x1c_parent_task_run_external_idB\x0e\n\x0c_child_indexB\x0c\n\n_child_keyB\x16\n\x14_additional_metadataB\x14\n\x12_desired_worker_idB\x0b\n\t_priority*\x85\x01\n\x15WorkerLabelComparator\x12\t\n\x05\x45QUAL\x10\x00\x12\r\n\tNOT_EQUAL\x10\x01\x12\x10\n\x0cGREATER_THAN\x10\x02\x12\x19\n\x15GREATER_THAN_OR_EQUAL\x10\x03\x12\r\n\tLESS_THAN\x10\x04\x12\x16\n\x12LESS_THAN_OR_EQUAL\x10\x05\x42\x42Z@github.com/hatchet-dev/hatchet/internal/services/shared/proto/v1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'v1.shared.trigger_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z@github.com/hatchet-dev/hatchet/internal/services/shared/proto/v1'
  _globals['_TRIGGERWORKFLOWREQUEST_DESIREDWORKERLABELSENTRY']._loaded_options = None
  _globals['_TRIGGERWORKFLOWREQUEST_DESIREDWORKERLABELSENTRY']._serialized_options = b'8\001'
  _globals['_WORKERLABELCOMPARATOR']._serialized_start=832
  _globals['_WORKERLABELCOMPARATOR']._serialized_end=965
  _globals['_DESIREDWORKERLABELS']._serialized_start=32
  _globals['_DESIREDWORKERLABELS']._serialized_end=264
  _globals['_TRIGGERWORKFLOWREQUEST']._serialized_start=267
  _globals['_TRIGGERWORKFLOWREQUEST']._serialized_end=829
  _globals['_TRIGGERWORKFLOWREQUEST_DESIREDWORKERLABELSENTRY']._serialized_start=611
  _globals['_TRIGGERWORKFLOWREQUEST_DESIREDWORKERLABELSENTRY']._serialized_end=694
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager

from hatchet_sdk.runnables.action import ActionKey
from hatchet_sdk.worker.durable_eviction.manager import DurableEvictionManager


@asynccontextmanager
async def aio_durable_eviction_wait(
    wait_kind: str,
    resource_id: str,
    action_key: ActionKey | None = None,
    eviction_manager: DurableEvictionManager | None = None,
) -> AsyncIterator[None]:
    """
    Mark an SDK-managed wait for the current durable run (if applicable).

    If action_key or eviction_manager is None, this is a no-op.
    """

    if action_key and eviction_manager is not None:
        eviction_manager.mark_waiting(
            action_key, wait_kind=wait_kind, resource_id=resource_id
        )

    try:
        yield
    finally:
        if action_key and eviction_manager is not None:
            eviction_manager.mark_active(action_key)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: v1/dispatcher.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'v1/dispatcher.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from hatchet_sdk.contracts.v1.shared import condition_pb2 as v1_dot_shared_dot_condition__pb2
from hatchet_sdk.contracts.v1.shared import trigger_pb2 as v1_dot_shared_dot_trigger__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13v1/dispatcher.proto\x12\x02v1\x1a\x19v1/shared/condition.proto\x1a\x17v1/shared/trigger.proto\"5\n DurableTaskRequestRegisterWorker\x12\x11\n\tworker_id\x18\x01 \x01(\t\"6\n!DurableTaskResponseRegisterWorker\x12\x11\n\tworker_id\x18\x01 \x01(\t\"y\n\x17\x44urableEventLogEntryRef\x12 \n\x18\x64urable_task_external_id\x18\x01 \x01(\t\x12\x18\n\x10invocation_count\x18\x02 \x01(\x05\x12\x11\n\tbranch_id\x18\x03 \x01(\x03\x12\x0f\n\x07node_id\x18\x04 \x01(\x03\"<\n\x16\x44urableTaskRunAckEntry\x12\x0f\n\x07node_id\x18\x01 \x01(\x03\x12\x11\n\tbranch_id\x18\x02 \x01(\x03\"\xa3\x01\n\x1f\x44urableTaskEventMemoAckResponse\x12(\n\x03ref\x18\x01 \x01(\x0b\x32\x1b.v1.DurableEventLogEntryRef\x12\x1c\n\x14memo_already_existed\x18\x02 \x01(\x08\x12 \n\x13memo_result_payload\x18\x03 \x01(\x0cH\x00\x88\x01\x01\x42\x16\n\x14_memo_result_payload\"\x95\x01\n&DurableTaskEventTriggerRunsAckResponse\x12 \n\x18\x64urable_task_external_id\x18\x01 \x01(\t\x12\x18\n\x10invocation_count\x18\x02 \x01(\x05\x12/\n\x0brun_entries\x18\x03 \x03(\x0b\x32\x1a.v1.DurableTaskRunAckEntry\"N\n\"DurableTaskEventWaitForAckResponse\x12(\n\x03ref\x18\x01 \x01(\x0b\x32\x1b.v1.DurableEventLogEntryRef\"f\n)DurableTaskEventLogEntryCompletedResponse\x12(\n\x03ref\x18\x01 \x01(\x0b\x32\x1b.v1.DurableEventLogEntryRef\x12\x0f\n\x07payload\x18\x02 \x01(\x0c\"\x7f\n!DurableTaskEvictInvocationRequest\x12\x18\n\x10invocation_count\x18\x01 \x01(\x05\x12 \n\x18\x64urable_task_external_id\x18\x02 \x01(\t\x12\x13\n\x06reason\x18\x03 \x01(\tH\x00\x88\x01\x01\x42\t\n\x07_reason\"\\\n\x1e\x44urableTaskEvictionAckResponse\x12\x18\n\x10invocation_count\x18\x01 \x01(\x05\x12 \n\x18\x64urable_task_external_id\x18\x02 \x01(\t\"\x82\x01\n DurableTaskAwaitedCompletedEntry\x12 \n\x18\x64urable_task_external_id\x18\x01 \x01(\t\x12\x11\n\tbranch_id\x18\x02 \x01(\x03\x12\x0f\n\x07node_id\x18\x03 \x01(\x03\x12\x18\n\x10invocation_count\x18\x04 \x01(\x05\"j\n\x1c\x44urableTaskServerEvictNotice\x12 \n\x18\x64urable_task_external_id\x18\x01 \x01(\t\x12\x18\n\x10invocation_count\x18\x02 \x01(\x05\x12\x0e\n\x06reason\x18\x03 \x01(\t\"r\n\x1e\x44urableTaskWorkerStatusRequest\x12\x11\n\tworker_id\x18\x01 \x01(\t\x12=\n\x0fwaiting_entries\x18\x02 \x03(\x0b\x32$.v1.DurableTaskAwaitedCompletedEntry\"m\n\x1e\x44urableTaskCompleteMemoRequest\x12(\n\x03ref\x18\x01 \x01(\x0b\x32\x1b.v1.DurableEventLogEntryRef\x12\x0f\n\x07payload\x18\x02 \x01(\x0c\x12\x10\n\x08memo_key\x18\x03 \x01(\x0c\"\x83\x01\n\x16\x44urableTaskMemoRequest\x12\x18\n\x10invocation_count\x18\x01 \x01(\x05\x12 \n\x18\x64urable_task_external_id\x18\x02 \x01(\t\x12\x0b\n\x03key\x18\x03 \x01(\x0c\x12\x14\n\x07payload\x18\x04 \x01(\x0cH\x00\x88\x01\x01\x42\n\n\x08_payload\"\x8d\x01\n\x1d\x44urableTaskTriggerRunsRequest\x12\x18\n\x10invocation_count\x18\x01 \x01(\x05\x12 \n\x18\x64urable_task_external_id\x18\x02 \x01(\t\x12\x30\n\x0ctrigger_opts\x18\x03 \x03(\x0b\x32\x1a.v1.TriggerWorkflowRequest\"\xb5\x01\n\x19\x44urableTaskWaitForRequest\x12\x18\n\x10invocation_count\x18\x01 \x01(\x05\x12 \n\x18\x64urable_task_external_id\x18\x02 \x01(\t\x12\x44\n\x13wait_for_conditions\x18\x03 \x01(\x0b\x32\".v1.DurableEventListenerConditionsH\x00\x88\x01\x01\x42\x16\n\x14_wait_for_conditions\"\xb7\x03\n\x12\x44urableTaskRequest\x12?\n\x0fregister_worker\x18\x01 \x01(\x0b\x32$.v1.DurableTaskRequestRegisterWorkerH\x00\x12*\n\x04memo\x18\x02 \x01(\x0b\x32\x1a.v1.DurableTaskMemoRequestH\x00\x12\x39\n\x0ctrigger_runs\x18\x03 \x01(\x0b\x32!.v1.DurableTaskTriggerRunsRequestH\x00\x12\x31\n\x08wait_for\x18\x04 \x01(\x0b\x32\x1d.v1.DurableTaskWaitForRequestH\x00\x12\x41\n\x10\x65vict_invocation\x18\x05 \x01(\x0b\x32%.v1.DurableTaskEvictInvocationRequestH\x00\x12;\n\rworker_status\x18\x06 \x01(\x0b\x32\".v1.DurableTaskWorkerStatusRequestH\x00\x12;\n\rcomplete_memo\x18\x07 \x01(\x0b\x32\".v1.DurableTaskCompleteMemoRequestH\x00\x42\t\n\x07message\"\x89\x01\n\x18\x44urableTaskErrorResponse\x12(\n\x03ref\x18\x01 \x01(\x0b\x32\x1b.v1.DurableEventLogEntryRef\x12,\n\nerror_type\x18\x02 \x01(\x0e\x32\x18.v1.DurableTaskErrorType\x12\x15\n\rerror_message\x18\x03 \x01(\t\"\x92\x04\n\x13\x44urableTaskResponse\x12@\n\x0fregister_worker\x18\x01 \x01(\x0b\x32%.v1.DurableTaskResponseRegisterWorkerH\x00\x12\x37\n\x08memo_ack\x18\x02 \x01(\x0b\x32#.v1.DurableTaskEventMemoAckResponseH\x00\x12\x46\n\x10trigger_runs_ack\x18\x03 \x01(\x0b\x32*.v1.DurableTaskEventTriggerRunsAckResponseH\x00\x12>\n\x0cwait_for_ack\x18\x04 \x01(\x0b\x32&.v1.DurableTaskEventWaitForAckResponseH\x00\x12H\n\x0f\x65ntry_completed\x18\x05 \x01(\x0b\x32-.v1.DurableTaskEventLogEntryCompletedResponseH\x00\x12-\n\x05\x65rror\x18\x06 \x01(\x0b\x32\x1c.v1.DurableTaskErrorResponseH\x00\x12:\n\x0c\x65viction_ack\x18\x07 \x01(\x0b\x32\".v1.DurableTaskEvictionAckResponseH\x00\x12\x38\n\x0cserver_evict\x18\x08 \x01(\x0b\x32 .v1.DurableTaskServerEvictNoticeH\x00\x42\t\n\x07message\"z\n\x1bRegisterDurableEventRequest\x12\x0f\n\x07task_id\x18\x01 \x01(\t\x12\x12\n\nsignal_key\x18\x02 \x01(\t\x12\x36\n\nconditions\x18\x03 \x01(\x0b\x32\".v1.DurableEventListenerConditions\"\x1e\n\x1cRegisterDurableEventResponse\"C\n\x1cListenForDurableEventRequest\x12\x0f\n\x07task_id\x18\x01 \x01(\t\x12\x12\n\nsignal_key\x18\x02 \x01(\t\"A\n\x0c\x44urableEvent\x12\x0f\n\x07task_id\x18\x01 \x01(\t\x12\x12\n\nsignal_key\x18\x02 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c*k\n\x14\x44urableTaskErrorType\x12\'\n#DURABLE_TASK_ERROR_TYPE_UNSPECIFIED\x10\x00\x12*\n&DURABLE_TASK_ERROR_TYPE_NONDETERMINISM\x10\x01\x32\x84\x02\n\x0cV1Dispatcher\x12\x44\n\x0b\x44urableTask\x12\x16.v1.DurableTaskRequest\x1a\x17.v1.DurableTaskResponse\"\x00(\x01\x30\x01\x12[\n\x14RegisterDurableEvent\x12\x1f.v1.RegisterDurableEventRequest\x1a .v1.RegisterDurableEventResponse\"\x00\x12Q\n\x15ListenForDurableEvent\x12 .v1.ListenForDurableEventRequest\x1a\x10.v1.DurableEvent\"\x00(\x01\x30\x01\x42\x42Z@github.com/hatchet-dev/hatchet/internal/services/shared/proto/v1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'v1.dispatcher_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z@github.com/hatchet-dev/hatchet/internal/services/shared/proto/v1'
  _globals['_DURABLETASKERRORTYPE']._serialized_start=3437
  _globals['_DURABLETASKERRORTYPE']._serialized_end=3544
  _globals['_DURABLETASKREQUESTREGISTERWORKER']._serialized_start=79
  _globals['_DURABLETASKREQUESTREGISTERWORKER']._serialized_end=132
  _globals['_DURABLETASKRESPONSEREGISTERWORKER']._serialized_start=134
  _globals['_DURABLETASKRESPONSEREGISTERWORKER']._serialized_end=188
  _globals['_DURABLEEVENTLOGENTRYREF']._serialized_start=190
  _globals['_DURABLEEVENTLOGENTRYREF']._serialized_end=311
  _globals['_DURABLETASKRUNACKENTRY']._serialized_start=313
  _globals['_DURABLETASKRUNACKENTRY']._serialized_end=373
  _globals['_DURABLETASKEVENTMEMOACKRESPONSE']._serialized_start=376
  _globals['_DURABLETASKEVENTMEMOACKRESPONSE']._serialized_end=539
  _globals['_DURABLETASKEVENTTRIGGERRUNSACKRESPONSE']._serialized_start=542
  _globals['_DURABLETASKEVENTTRIGGERRUNSACKRESPONSE']._serialized_end=691
  _globals['_DURABLETASKEVENTWAITFORACKRESPONSE']._serialized_start=693
  _globals['_DURABLETASKEVENTWAITFORACKRESPONSE']._serialized_end=771
  _globals['_DURABLETASKEVENTLOGENTRYCOMPLETEDRESPONSE']._serialized_start=773
  _globals['_DURABLETASKEVENTLOGENTRYCOMPLETEDRESPONSE']._serialized_end=875
  _globals['_DURABLETASKEVICTINVOCATIONREQUEST']._serialized_start=877
  _globals['_DURABLETASKEVICTINVOCATIONREQUEST']._serialized_end=1004
  _globals['_DURABLETASKEVICTIONACKRESPONSE']._serialized_start=1006
  _globals['_DURABLETASKEVICTIONACKRESPONSE']._serialized_end=1098
  _globals['_DURABLETASKAWAITEDCOMPLETEDENTRY']._serialized_start=1101
  _globals['_DURABLETASKAWAITEDCOMPLETEDENTRY']._serialized_end=1231
  _globals['_DURABLETASKSERVEREVICTNOTICE']._serialized_start=1233
  _globals['_DURABLETASKSERVEREVICTNOTICE']._serialized_end=1339
  _globals['_DURABLETASKWORKERSTATUSREQUEST']._serialized_start=1341
  _globals['_DURABLETASKWORKERSTATUSREQUEST']._serialized_end=1455
  _globals['_DURABLETASKCOMPLETEMEMOREQUEST']._serialized_start=1457
  _globals['_DURABLETASKCOMPLETEMEMOREQUEST']._serialized_end=1566
  _globals['_DURABLETASKMEMOREQUEST']._serialized_start=1569
  _globals['_DURABLETASKMEMOREQUEST']._serialized_end=1700
  _globals['_DURABLETASKTRIGGERRUNSREQUEST']._serialized_start=1703
  _globals['_DURABLETASKTRIGGERRUNSREQUEST']._serialized_end=1844
  _globals['_DURABLETASKWAITFORREQUEST']._serialized_start=1847
  _globals['_DURABLETASKWAITFORREQUEST']._serialized_end=2028
  _globals['_DURABLETASKREQUEST']._serialized_start=2031
  _globals['_DURABLETASKREQUEST']._serialized_end=2470
  _globals['_DURABLETASKERRORRESPONSE']._serialized_start=2473
  _globals['_DURABLETASKERRORRESPONSE']._serialized_end=2610
  _globals['_DURABLETASKRESPONSE']._serialized_start=2613
  _globals['_DURABLETASKRESPONSE']._serialized_end=3143
  _globals['_REGISTERDURABLEEVENTREQUEST']._serialized_start=3145
  _globals['_REGISTERDURABLEEVENTREQUEST']._serialized_end=3267
  _globals['_REGISTERDURABLEEVENTRESPONSE']._serialized_start=3269
  _globals['_REGISTERDURABLEEVENTRESPONSE']._serialized_end=3299
  _globals['_LISTENFORDURABLEEVENTREQUEST']._serialized_start=3301
  _globals['_LISTENFORDURABLEEVENTREQUEST']._serialized_end=3368
  _globals['_DURABLEEVENT']._serialized_start=3370
  _globals['_DURABLEEVENT']._serialized_end=3435
  _globals['_V1DISPATCHER']._serialized_start=3547
  _globals['_V1DISPATCHER']._serialized_end=3807
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from datetime import datetime, timedelta
from enum import Enum

from pydantic import BaseModel
from typing_extensions import assert_never

from hatchet_sdk.logger import logger
from hatchet_sdk.runnables.action import ActionKey
from hatchet_sdk.runnables.eviction import EvictionPolicy


class EvictionCause(str, Enum):
    TTL_EXCEEDED = "ttl_exceeded"
    CAPACITY_PRESSURE = "capacity_pressure"
    WORKER_SHUTDOWN = "worker_shutdown"


class DurableRunRecord(BaseModel):
    key: ActionKey
    step_run_id: str
    invocation_count: int
    eviction_policy: EvictionPolicy | None
    registered_at: datetime

    # Waiting state -- ref-counted so concurrent waits (e.g. asyncio.gather
    # over multiple child results) don't prematurely clear the waiting flag
    # when one child completes before the others.
    waiting_since: datetime | None = None
    wait_kind: str | None = None
    wait_resource_id: str | None = None
    _wait_count: int = 0

    # Set by the eviction manager before requesting eviction
    eviction_reason: str | None = None

    @property
    def is_waiting(self) -> bool:
        return self._wait_count > 0


class DurableEvictionCache:
    def __init__(self) -> None:
        self._runs: dict[ActionKey, DurableRunRecord] = {}

    def register_run(
        self,
        key: ActionKey,
        step_run_id: str,
        invocation_count: int,
        now: datetime,
        eviction_policy: EvictionPolicy | None,
    ) -> None:
        self._runs[key] = DurableRunRecord(
            key=key,
            step_run_id=step_run_id,
            invocation_count=invocation_count,
            eviction_policy=eviction_policy,
            registered_at=now,
        )

    def unregister_run(self, key: ActionKey) -> None:
        self._runs.pop(key, None)

    def get(self, key: ActionKey) -> DurableRunRecord | None:
        return self._runs.get(key)

    def get_all_waiting(self) -> list[DurableRunRecord]:
        return [r for r in self._runs.values() if r.is_waiting]

    def find_key_by_step_run_id(self, step_run_id: str) -> ActionKey | None:
        for key, rec in self._runs.items():
            if rec.step_run_id == step_run_id:
                return key
        return None

    def mark_waiting(
        self,
        key: ActionKey,
        now: datetime,
        wait_kind: str,
        resource_id: str,
    ) -> None:
        rec = self._runs.get(key)
        if not rec:
            return

        rec._wait_count += 1
        if rec._wait_count == 1:
            rec.waiting_since = now
        rec.wait_kind = wait_kind
        rec.wait_resource_id = resource_id

    def mark_active(self, key: ActionKey, now: datetime) -> None:
        rec = self._runs.get(key)
        if not rec:
            return

        rec._wait_count = max(0, rec._wait_count - 1)
        if rec._wait_count == 0:
            rec.waiting_since = None
            rec.wait_kind = None
            rec.wait_resource_id = None

    def _capacity_pressure(
        self, durable_slots: int, reserve_slots: int, waiting_count: int
    ) -> bool:
        if durable_slots <= 0:
            return False

        max_waiting = durable_slots - reserve_slots
        if max_waiting <= 0:
            return False

        return waiting_count >= max_waiting

    def select_eviction_candidate(
        self,
        now: datetime,
        durable_slots: int,
        reserve_slots: int,
        min_wait_for_capacity_eviction: timedelta,
    ) -> ActionKey | None:
        waiting: list[DurableRunRecord] = [
            r
            for r in self._runs.values()
            if r.is_waiting and r.eviction_policy is not None
        ]

        if not waiting:
            return None

        # Prefer TTL-eligible candidates first.
        ttl_eligible: list[DurableRunRecord] = [
            r
            for r in waiting
            if r.eviction_policy is not None
            and r.eviction_policy.ttl is not None
            and r.waiting_since is not None
            and (now - r.waiting_since) >= r.eviction_policy.ttl
        ]

        if ttl_eligible:
            ttl_eligible.sort(
                key=lambda r: (
                    r.eviction_policy.priority if r.eviction_policy else 0,
                    r.waiting_since or now,
                )
            )
            chosen = ttl_eligible[0]
            ttl = chosen.eviction_policy.ttl if chosen.eviction_policy else None
            chosen.eviction_reason = _build_eviction_reason(
                EvictionCause.TTL_EXCEEDED, chosen, ttl=ttl
            )
            logger.debug(
                "DurableEvictionCache: TTL eviction candidate selected "
                f"step_run_id={chosen.step_run_id} kind={chosen.wait_kind}"
            )
            return chosen.key

        # Capacity eviction: only if we're above waiting capacity and run allows it.
        capacity_pressure = self._capacity_pressure(
            durable_slots=durable_slots,
            reserve_slots=reserve_slots,
            waiting_count=len(waiting),
        )
        if not capacity_pressure:
            return None

        capacity_candidates: list[DurableRunRecord] = [
            r
            for r in waiting
            if r.eviction_policy
            and r.eviction_policy.allow_capacity_eviction
            and r.waiting_since is not None
            and (now - r.waiting_since) >= min_wait_for_capacity_eviction
        ]

        if not capacity_candidates:
            return None

        capacity_candidates.sort(
            key=lambda r: (
                r.eviction_policy.priority if r.eviction_policy else 0,
                r.waiting_since or now,
            )
        )
        chosen = capacity_candidates[0]
        chosen.eviction_reason = _build_eviction_reason(
            EvictionCause.CAPACITY_PRESSURE, chosen
        )
        logger.debug(
            "DurableEvictionCache: capacity eviction candidate selected "
            f"step_run_id={chosen.step_run_id} kind={chosen.wait_kind}"
        )
        return chosen.key


def _build_eviction_reason(
    cause: EvictionCause,
    rec: DurableRunRecord,
    ttl: timedelta | None = None,
) -> str:
    wait_desc = rec.wait_kind or "unknown"
    if rec.wait_resource_id:
        wait_desc = f"{wait_desc}({rec.wait_resource_id})"

    match cause:
        case EvictionCause.TTL_EXCEEDED:
            ttl_str = f" ({ttl})" if ttl else ""
            return f"Wait TTL{ttl_str} exceeded while waiting on {wait_desc}"
        case EvictionCause.CAPACITY_PRESSURE:
            return f"Worker at capacity while waiting on {wait_desc}"
        case EvictionCause.WORKER_SHUTDOWN:
            return f"Worker shutdown while waiting on {wait_desc}"
        case _ as unreachable:
            assert_never(unreachable)



from __future__ import annotations

from datetime import timedelta

from pydantic import BaseModel, ConfigDict


class EvictionPolicy(BaseModel):
    """
    Task-scoped eviction parameters for *durable* tasks.

    :ivar ttl: Maximum continuous waiting duration before TTL-eligible eviction.
        Applies to time spent in SDK-instrumented "waiting" states (e.g.
        `ctx.aio_wait_for(...)`, waiting for a workflow run result).
    :ivar allow_capacity_eviction: Whether this task may be evicted under durable-slot pressure.
    :ivar priority: Lower values are evicted first when multiple candidates exist.

    Setting the durable task's eviction params to `None` means the task run is
    never eligible for eviction.

    **Example**
    ```python
    EvictionPolicy(
        ttl=timedelta(minutes=10),
        allow_capacity_eviction=True,
        priority=0,
    )
    ```
    """

    model_config = ConfigDict(frozen=True)

    ttl: timedelta | None

    allow_capacity_eviction: bool = True

    priority: int = 0


# Shared sensible defaults (single source of truth).
# NOTE: When changing these values, update the :param durable_run_eviction: / :param eviction_policy:
# docstrings in workflow.Workflow.durable_task and hatchet.Hatchet.durable_task to match.
DEFAULT_DURABLE_TASK_EVICTION_POLICY = EvictionPolicy(
    ttl=timedelta(minutes=15),
    allow_capacity_eviction=True,
    priority=0,
)



from __future__ import annotations

import asyncio
from collections.abc import Awaitable, Callable
from datetime import datetime, timedelta, timezone

from pydantic import BaseModel, ConfigDict

from hatchet_sdk.logger import logger
from hatchet_sdk.runnables.action import ActionKey
from hatchet_sdk.runnables.eviction import EvictionPolicy
from hatchet_sdk.worker.durable_eviction.cache import (
    DurableEvictionCache,
    DurableRunRecord,
    EvictionCause,
    _build_eviction_reason,
)


class DurableEvictionConfig(BaseModel):
    model_config = ConfigDict(frozen=True)

    check_interval: timedelta = timedelta(seconds=1)
    """How often we try selecting an eviction candidate."""

    reserve_slots: int = 0
    """How many slots to reserve from capacity-based eviction decisions."""

    min_wait_for_capacity_eviction: timedelta = timedelta(seconds=10)
    """Avoid immediately evicting runs that have just entered a wait."""


DEFAULT_DURABLE_EVICTION_CONFIG = DurableEvictionConfig()


class DurableEvictionManager:
    def __init__(
        self,
        *,
        durable_slots: int,
        cancel_local: Callable[[ActionKey], None],
        request_eviction_with_ack: Callable[
            [ActionKey, DurableRunRecord], Awaitable[None]
        ],
        config: DurableEvictionConfig = DEFAULT_DURABLE_EVICTION_CONFIG,
        cache: DurableEvictionCache | None = None,
    ) -> None:
        self._durable_slots = durable_slots
        self._cancel_local = cancel_local
        self._request_eviction_with_ack = request_eviction_with_ack
        self._config = config
        self._cache = cache or DurableEvictionCache()

        self._task: asyncio.Task[None] | None = None
        self._lock = asyncio.Lock()

    @property
    def cache(self) -> DurableEvictionCache:
        return self._cache

    def start(self) -> None:
        # Lazy start (requires an event loop)
        if self._task and not self._task.done():
            return
        self._task = asyncio.create_task(self._run_loop())

    def stop(self) -> None:
        if self._task and not self._task.done():
            self._task.cancel()

    def register_run(
        self,
        key: ActionKey,
        *,
        step_run_id: str,
        invocation_count: int,
        eviction_policy: EvictionPolicy | None,
    ) -> None:
        self._cache.register_run(
            key,
            step_run_id,
            invocation_count=invocation_count,
            now=self._now(),
            eviction_policy=eviction_policy,
        )

    def unregister_run(self, key: ActionKey) -> None:
        self._cache.unregister_run(key)

    def mark_waiting(
        self,
        key: ActionKey,
        *,
        wait_kind: str,
        resource_id: str,
    ) -> None:
        self._cache.mark_waiting(
            key,
            now=self._now(),
            wait_kind=wait_kind,
            resource_id=resource_id,
        )

    def mark_active(self, key: ActionKey) -> None:
        self._cache.mark_active(key, now=self._now())

    def _evict_run(self, key: ActionKey) -> None:
        self._cancel_local(key)
        self.unregister_run(key)

    async def _run_loop(self) -> None:
        interval = self._config.check_interval.total_seconds()

        try:
            while True:
                await asyncio.sleep(interval)
                await self._tick_safe()
        except asyncio.CancelledError:
            return

    async def _tick_safe(self) -> None:
        try:
            await self._tick()
        except Exception:
            logger.exception("DurableEvictionManager: error in eviction loop")

    async def _tick(self) -> None:
        # Only one eviction *cycle* at a time.
        #
        # Within a tick we drain all currently-eligible candidates
        async with self._lock:
            evicted_this_tick: set[ActionKey] = set()

            while True:
                key = self._cache.select_eviction_candidate(
                    now=self._now(),
                    durable_slots=self._durable_slots,
                    reserve_slots=self._config.reserve_slots,
                    min_wait_for_capacity_eviction=self._config.min_wait_for_capacity_eviction,
                )
                if key is None:
                    return

                # Safety: avoid infinite loops if cache repeatedly returns same key.
                if key in evicted_this_tick:
                    return
                evicted_this_tick.add(key)

                rec = self._cache.get(key)
                if rec is None:
                    continue

                if rec.eviction_policy is None:
                    continue

                logger.debug(
                    "DurableEvictionManager: evicting durable run "
                    f"task_run_external_id={rec.step_run_id} wait_kind={rec.wait_kind} "
                    f"resource_id={rec.wait_resource_id} ttl={rec.eviction_policy.ttl} "
                    f"capacity_allowed={rec.eviction_policy.allow_capacity_eviction}"
                )

                await self._request_eviction_with_ack(key, rec)
                self._evict_run(key)

    def handle_server_eviction(self, step_run_id: str, invocation_count: int) -> None:
        """Handle a server-initiated eviction notification for a stale invocation."""
        key = self._cache.find_key_by_step_run_id(step_run_id)
        if key is None:
            return

        rec = self._cache.get(key)
        if rec is not None and rec.invocation_count != invocation_count:
            return

        logger.info(
            "DurableEvictionManager: server-initiated eviction for "
            "step_run_id=%s invocation_count=%d",
            step_run_id,
            invocation_count,
        )
        self._evict_run(key)

    async def evict_all_waiting(self) -> int:
        """Evict every currently-waiting durable run. Used during graceful shutdown."""
        self.stop()

        waiting = self._cache.get_all_waiting()
        evicted = 0

        for rec in waiting:
            rec.eviction_reason = _build_eviction_reason(
                EvictionCause.WORKER_SHUTDOWN, rec
            )

            logger.debug(
                "DurableEvictionManager: shutdown-evicting durable run "
                f"task_run_external_id={rec.step_run_id} wait_kind={rec.wait_kind} "
                f"resource_id={rec.wait_resource_id}"
            )

            try:
                await self._request_eviction_with_ack(rec.key, rec)
            except Exception:
                logger.exception(
                    f"DurableEvictionManager: failed to send eviction for "
                    f"step_run_id={rec.step_run_id}"
                )

            # Always cancel locally even if the server ACK failed, so the
            # future settles and exit_gracefully doesn't hang.
            self._evict_run(rec.key)
            evicted += 1

        return evicted

    def _now(self) -> datetime:
        return datetime.now(timezone.utc)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: workflows.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'workflows.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from hatchet_sdk.contracts.v1.shared import trigger_pb2 as v1_dot_shared_dot_trigger__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0fworkflows.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x17v1/shared/trigger.proto\">\n\x12PutWorkflowRequest\x12(\n\x04opts\x18\x01 \x01(\x0b\x32\x1a.CreateWorkflowVersionOpts\"\xbf\x04\n\x19\x43reateWorkflowVersionOpts\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\t\x12\x16\n\x0e\x65vent_triggers\x18\x04 \x03(\t\x12\x15\n\rcron_triggers\x18\x05 \x03(\t\x12\x36\n\x12scheduled_triggers\x18\x06 \x03(\x0b\x32\x1a.google.protobuf.Timestamp\x12$\n\x04jobs\x18\x07 \x03(\x0b\x32\x16.CreateWorkflowJobOpts\x12-\n\x0b\x63oncurrency\x18\x08 \x01(\x0b\x32\x18.WorkflowConcurrencyOpts\x12\x1d\n\x10schedule_timeout\x18\t \x01(\tH\x00\x88\x01\x01\x12\x17\n\ncron_input\x18\n \x01(\tH\x01\x88\x01\x01\x12\x33\n\x0eon_failure_job\x18\x0b \x01(\x0b\x32\x16.CreateWorkflowJobOptsH\x02\x88\x01\x01\x12$\n\x06sticky\x18\x0c \x01(\x0e\x32\x0f.StickyStrategyH\x03\x88\x01\x01\x12 \n\x04kind\x18\r \x01(\x0e\x32\r.WorkflowKindH\x04\x88\x01\x01\x12\x1d\n\x10\x64\x65\x66\x61ult_priority\x18\x0e \x01(\x05H\x05\x88\x01\x01\x42\x13\n\x11_schedule_timeoutB\r\n\x0b_cron_inputB\x11\n\x0f_on_failure_jobB\t\n\x07_stickyB\x07\n\x05_kindB\x13\n\x11_default_priority\"\xd0\x01\n\x17WorkflowConcurrencyOpts\x12\x13\n\x06\x61\x63tion\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x15\n\x08max_runs\x18\x02 \x01(\x05H\x01\x88\x01\x01\x12\x36\n\x0elimit_strategy\x18\x03 \x01(\x0e\x32\x19.ConcurrencyLimitStrategyH\x02\x88\x01\x01\x12\x17\n\nexpression\x18\x04 \x01(\tH\x03\x88\x01\x01\x42\t\n\x07_actionB\x0b\n\t_max_runsB\x11\n\x0f_limit_strategyB\r\n\x0b_expression\"h\n\x15\x43reateWorkflowJobOpts\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12&\n\x05steps\x18\x04 \x03(\x0b\x32\x17.CreateWorkflowStepOptsJ\x04\x08\x03\x10\x04\"\xb8\x03\n\x16\x43reateWorkflowStepOpts\x12\x13\n\x0breadable_id\x18\x01 \x01(\t\x12\x0e\n\x06\x61\x63tion\x18\x02 \x01(\t\x12\x0f\n\x07timeout\x18\x03 \x01(\t\x12\x0e\n\x06inputs\x18\x04 \x01(\t\x12\x0f\n\x07parents\x18\x05 \x03(\t\x12\x11\n\tuser_data\x18\x06 \x01(\t\x12\x0f\n\x07retries\x18\x07 \x01(\x05\x12)\n\x0brate_limits\x18\x08 \x03(\x0b\x32\x14.CreateStepRateLimit\x12@\n\rworker_labels\x18\t \x03(\x0b\x32).CreateWorkflowStepOpts.WorkerLabelsEntry\x12\x1b\n\x0e\x62\x61\x63koff_factor\x18\n \x01(\x02H\x00\x88\x01\x01\x12 \n\x13\x62\x61\x63koff_max_seconds\x18\x0b \x01(\x05H\x01\x88\x01\x01\x1aL\n\x11WorkerLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12&\n\x05value\x18\x02 \x01(\x0b\x32\x17.v1.DesiredWorkerLabels:\x02\x38\x01\x42\x11\n\x0f_backoff_factorB\x16\n\x14_backoff_max_seconds\"\xfa\x01\n\x13\x43reateStepRateLimit\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x12\n\x05units\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x15\n\x08key_expr\x18\x03 \x01(\tH\x01\x88\x01\x01\x12\x17\n\nunits_expr\x18\x04 \x01(\tH\x02\x88\x01\x01\x12\x1e\n\x11limit_values_expr\x18\x05 \x01(\tH\x03\x88\x01\x01\x12)\n\x08\x64uration\x18\x06 \x01(\x0e\x32\x12.RateLimitDurationH\x04\x88\x01\x01\x42\x08\n\x06_unitsB\x0b\n\t_key_exprB\r\n\x0b_units_exprB\x14\n\x12_limit_values_exprB\x0b\n\t_duration\"\x16\n\x14ListWorkflowsRequest\"\x83\x03\n\x17ScheduleWorkflowRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12-\n\tschedules\x18\x02 \x03(\x0b\x32\x1a.google.protobuf.Timestamp\x12\r\n\x05input\x18\x03 \x01(\t\x12\x16\n\tparent_id\x18\x04 \x01(\tH\x00\x88\x01\x01\x12(\n\x1bparent_task_run_external_id\x18\x05 \x01(\tH\x01\x88\x01\x01\x12\x18\n\x0b\x63hild_index\x18\x06 \x01(\x05H\x02\x88\x01\x01\x12\x16\n\tchild_key\x18\x07 \x01(\tH\x03\x88\x01\x01\x12 \n\x13\x61\x64\x64itional_metadata\x18\x08 \x01(\tH\x04\x88\x01\x01\x12\x15\n\x08priority\x18\t \x01(\x05H\x05\x88\x01\x01\x42\x0c\n\n_parent_idB\x1e\n\x1c_parent_task_run_external_idB\x0e\n\x0c_child_indexB\x0c\n\n_child_keyB\x16\n\x14_additional_metadataB\x0b\n\t_priority\"O\n\x11ScheduledWorkflow\x12\n\n\x02id\x18\x01 \x01(\t\x12.\n\ntrigger_at\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\xe3\x01\n\x0fWorkflowVersion\x12\n\n\x02id\x18\x01 \x01(\t\x12.\n\ncreated_at\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nupdated_at\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0f\n\x07version\x18\x05 \x01(\t\x12\r\n\x05order\x18\x06 \x01(\x03\x12\x13\n\x0bworkflow_id\x18\x07 \x01(\t\x12/\n\x13scheduled_workflows\x18\x08 \x03(\x0b\x32\x12.ScheduledWorkflow\"?\n\x17WorkflowTriggerEventRef\x12\x11\n\tparent_id\x18\x01 \x01(\t\x12\x11\n\tevent_key\x18\x02 \x01(\t\"9\n\x16WorkflowTriggerCronRef\x12\x11\n\tparent_id\x18\x01 \x01(\t\x12\x0c\n\x04\x63ron\x18\x02 \x01(\t\"K\n\x1a\x42ulkTriggerWorkflowRequest\x12-\n\tworkflows\x18\x01 \x03(\x0b\x32\x1a.v1.TriggerWorkflowRequest\"7\n\x1b\x42ulkTriggerWorkflowResponse\x12\x18\n\x10workflow_run_ids\x18\x01 \x03(\t\"2\n\x17TriggerWorkflowResponse\x12\x17\n\x0fworkflow_run_id\x18\x01 \x01(\t\"W\n\x13PutRateLimitRequest\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05limit\x18\x02 \x01(\x05\x12$\n\x08\x64uration\x18\x03 \x01(\x0e\x32\x12.RateLimitDuration\"\x16\n\x14PutRateLimitResponse*$\n\x0eStickyStrategy\x12\x08\n\x04SOFT\x10\x00\x12\x08\n\x04HARD\x10\x01*2\n\x0cWorkflowKind\x12\x0c\n\x08\x46UNCTION\x10\x00\x12\x0b\n\x07\x44URABLE\x10\x01\x12\x07\n\x03\x44\x41G\x10\x02*\x7f\n\x18\x43oncurrencyLimitStrategy\x12\x16\n\x12\x43\x41NCEL_IN_PROGRESS\x10\x00\x12\x0f\n\x0b\x44ROP_NEWEST\x10\x01\x12\x10\n\x0cQUEUE_NEWEST\x10\x02\x12\x15\n\x11GROUP_ROUND_ROBIN\x10\x03\x12\x11\n\rCANCEL_NEWEST\x10\x04*]\n\x11RateLimitDuration\x12\n\n\x06SECOND\x10\x00\x12\n\n\x06MINUTE\x10\x01\x12\x08\n\x04HOUR\x10\x02\x12\x07\n\x03\x44\x41Y\x10\x03\x12\x08\n\x04WEEK\x10\x04\x12\t\n\x05MONTH\x10\x05\x12\x08\n\x04YEAR\x10\x06\x32\xdf\x02\n\x0fWorkflowService\x12\x34\n\x0bPutWorkflow\x12\x13.PutWorkflowRequest\x1a\x10.WorkflowVersion\x12>\n\x10ScheduleWorkflow\x12\x18.ScheduleWorkflowRequest\x1a\x10.WorkflowVersion\x12G\n\x0fTriggerWorkflow\x12\x1a.v1.TriggerWorkflowRequest\x1a\x18.TriggerWorkflowResponse\x12P\n\x13\x42ulkTriggerWorkflow\x12\x1b.BulkTriggerWorkflowRequest\x1a\x1c.BulkTriggerWorkflowResponse\x12;\n\x0cPutRateLimit\x12\x14.PutRateLimitRequest\x1a\x15.PutRateLimitResponseBBZ@github.com/hatchet-dev/hatchet/internal/services/admin/contractsb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'workflows_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z@github.com/hatchet-dev/hatchet/internal/services/admin/contracts'
  _globals['_CREATEWORKFLOWSTEPOPTS_WORKERLABELSENTRY']._loaded_options = None
  _globals['_CREATEWORKFLOWSTEPOPTS_WORKERLABELSENTRY']._serialized_options = b'8\001'
  _globals['_STICKYSTRATEGY']._serialized_start=2880
  _globals['_STICKYSTRATEGY']._serialized_end=2916
  _globals['_WORKFLOWKIND']._serialized_start=2918
  _globals['_WORKFLOWKIND']._serialized_end=2968
  _globals['_CONCURRENCYLIMITSTRATEGY']._serialized_start=2970
  _globals['_CONCURRENCYLIMITSTRATEGY']._serialized_end=3097
  _globals['_RATELIMITDURATION']._serialized_start=3099
  _globals['_RATELIMITDURATION']._serialized_end=3192
  _globals['_PUTWORKFLOWREQUEST']._serialized_start=77
  _globals['_PUTWORKFLOWREQUEST']._serialized_end=139
  _globals['_CREATEWORKFLOWVERSIONOPTS']._serialized_start=142
  _globals['_CREATEWORKFLOWVERSIONOPTS']._serialized_end=717
  _globals['_WORKFLOWCONCURRENCYOPTS']._serialized_start=720
  _globals['_WORKFLOWCONCURRENCYOPTS']._serialized_end=928
  _globals['_CREATEWORKFLOWJOBOPTS']._serialized_start=930
  _globals['_CREATEWORKFLOWJOBOPTS']._serialized_end=1034
  _globals['_CREATEWORKFLOWSTEPOPTS']._serialized_start=1037
  _globals['_CREATEWORKFLOWSTEPOPTS']._serialized_end=1477
  _globals['_CREATEWORKFLOWSTEPOPTS_WORKERLABELSENTRY']._serialized_start=1358
  _globals['_CREATEWORKFLOWSTEPOPTS_WORKERLABELSENTRY']._serialized_end=1434
  _globals['_CREATESTEPRATELIMIT']._serialized_start=1480
  _globals['_CREATESTEPRATELIMIT']._serialized_end=1730
  _globals['_LISTWORKFLOWSREQUEST']._serialized_start=1732
  _globals['_LISTWORKFLOWSREQUEST']._serialized_end=1754
  _globals['_SCHEDULEWORKFLOWREQUEST']._serialized_start=1757
  _globals['_SCHEDULEWORKFLOWREQUEST']._serialized_end=2144
  _globals['_SCHEDULEDWORKFLOW']._serialized_start=2146
  _globals['_SCHEDULEDWORKFLOW']._serialized_end=2225
  _globals['_WORKFLOWVERSION']._serialized_start=2228
  _globals['_WORKFLOWVERSION']._serialized_end=2455
  _globals['_WORKFLOWTRIGGEREVENTREF']._serialized_start=2457
  _globals['_WORKFLOWTRIGGEREVENTREF']._serialized_end=2520
  _globals['_WORKFLOWTRIGGERCRONREF']._serialized_start=2522
  _globals['_WORKFLOWTRIGGERCRONREF']._serialized_end=2579
  _globals['_BULKTRIGGERWORKFLOWREQUEST']._serialized_start=2581
  _globals['_BULKTRIGGERWORKFLOWREQUEST']._serialized_end=2656
  _globals['_BULKTRIGGERWORKFLOWRESPONSE']._serialized_start=2658
  _globals['_BULKTRIGGERWORKFLOWRESPONSE']._serialized_end=2713
  _globals['_TRIGGERWORKFLOWRESPONSE']._serialized_start=2715
  _globals['_TRIGGERWORKFLOWRESPONSE']._serialized_end=2765
  _globals['_PUTRATELIMITREQUEST']._serialized_start=2767
  _globals['_PUTRATELIMITREQUEST']._serialized_end=2854
  _globals['_PUTRATELIMITRESPONSE']._serialized_start=2856
  _globals['_PUTRATELIMITRESPONSE']._serialized_end=2878
  _globals['_WORKFLOWSERVICE']._serialized_start=3195
  _globals['_WORKFLOWSERVICE']._serialized_end=3546
# @@protoc_insertion_point(module_scope)

# Re-export for backwards compatibility
from hatchet_sdk.contracts.v1.shared.trigger_pb2 import TriggerWorkflowRequest as TriggerWorkflowRequest  # noqa: F401
from hatchet_sdk.contracts.v1.shared.trigger_pb2 import DesiredWorkerLabels as DesiredWorkerLabels  # noqa: F401
from hatchet_sdk.contracts.v1.shared.trigger_pb2 import WorkerLabelComparator as WorkerLabelComparator  # noqa: F401



from datetime import timedelta

HOUR = 3600
MINUTE = 60

Duration = timedelta | str


def timedelta_to_expr(td: Duration) -> str:
    if isinstance(td, str):
        return td

    ## `total_seconds` gives the entire duration,
    ## while `seconds` gives the seconds component of the timedelta
    ## e.g. 1 day and 1 second would give 86401 total seconds but 1 second
    seconds = int(td.total_seconds())

    ## IMPORTANT: We only support hours, minutes, and seconds on the engine
    if seconds % HOUR == 0:
        return f"{seconds // HOUR}h"
    if seconds % MINUTE == 0:
        return f"{seconds // MINUTE}m"
    return f"{seconds}s"


def expr_to_timedelta(expr: str) -> timedelta:
    unit = expr[-1]
    value = int(expr[:-1])

    if unit == "d":
        return timedelta(days=value)
    if unit == "h":
        return timedelta(hours=value)
    if unit == "m":
        return timedelta(minutes=value)
    if unit == "s":
        return timedelta(seconds=value)

    raise ValueError(f"Invalid time expression: {expr}")



import asyncio
import json
from collections.abc import AsyncIterator, Callable
from contextlib import suppress
from dataclasses import dataclass
from datetime import timedelta
from typing import Annotated, Literal, cast

import grpc.aio
from pydantic import BaseModel, Field
from typing_extensions import Never, Self

from hatchet_sdk.clients.admin import (
    AdminClient,
    TriggerWorkflowOptions,
)
from hatchet_sdk.config import ClientConfig
from hatchet_sdk.connection import new_conn
from hatchet_sdk.contracts.v1.dispatcher_pb2 import (
    DurableEventLogEntryRef,
    DurableTaskAwaitedCompletedEntry,
    DurableTaskCompleteMemoRequest,
    DurableTaskErrorType,
    DurableTaskEventLogEntryCompletedResponse,
    DurableTaskEvictInvocationRequest,
    DurableTaskMemoRequest,
    DurableTaskRequest,
    DurableTaskRequestRegisterWorker,
    DurableTaskResponse,
    DurableTaskTriggerRunsRequest,
    DurableTaskWaitForRequest,
    DurableTaskWorkerStatusRequest,
)
from hatchet_sdk.contracts.v1.dispatcher_pb2_grpc import V1DispatcherStub
from hatchet_sdk.contracts.v1.shared.condition_pb2 import DurableEventListenerConditions
from hatchet_sdk.exceptions import NonDeterminismError
from hatchet_sdk.logger import logger
from hatchet_sdk.metadata import get_metadata
from hatchet_sdk.utils.cache import TTLCache
from hatchet_sdk.utils.typing import JSONSerializableMapping

DEFAULT_RECONNECT_INTERVAL = 3  # seconds


@dataclass(frozen=True)
class WaitForEvent:
    wait_for_conditions: DurableEventListenerConditions


@dataclass(frozen=True)
class RunChildEvent:
    workflow_name: str
    input: str | None
    trigger_workflow_opts: TriggerWorkflowOptions


@dataclass(frozen=True)
class RunChildrenEvent:
    children: list[RunChildEvent]


@dataclass(frozen=True)
class MemoEvent:
    memo_key: bytes
    result: str | None


DurableTaskSendEvent = WaitForEvent | RunChildrenEvent | MemoEvent


class MaybeCachedMemoEntry(BaseModel):
    found: bool
    data: bytes | None = None


class DurableTaskRunAckEntry(BaseModel):
    node_id: int
    branch_id: int


class DurableTaskEventRunAck(BaseModel):
    ack_type: Literal["run"] = "run"
    invocation_count: int
    durable_task_external_id: str
    run_entries: list[DurableTaskRunAckEntry] = Field(default_factory=list)


class DurableTaskEventMemoAck(BaseModel):
    ack_type: Literal["memo"] = "memo"
    invocation_count: int
    durable_task_external_id: str
    branch_id: int
    node_id: int
    memo_already_existed: bool
    memo_result_payload: bytes | None = None


class DurableTaskEventWaitForAck(BaseModel):
    ack_type: Literal["wait"] = "wait"
    invocation_count: int
    durable_task_external_id: str
    branch_id: int
    node_id: int


DurableTaskEventAck = Annotated[
    DurableTaskEventRunAck | DurableTaskEventMemoAck | DurableTaskEventWaitForAck,
    Field(discriminator="ack_type"),
]


class DurableTaskEventLogEntryResult(BaseModel):
    durable_task_external_id: str
    node_id: int
    payload: JSONSerializableMapping | None

    @classmethod
    def from_proto(cls, proto: DurableTaskEventLogEntryCompletedResponse) -> Self:
        payload: JSONSerializableMapping | None = None
        if proto.payload:
            payload = json.loads(proto.payload.decode("utf-8"))

        return cls(
            durable_task_external_id=proto.ref.durable_task_external_id,
            node_id=proto.ref.node_id,
            payload=payload,
        )


TaskExternalId = str
NodeId = int
BranchId = int
InvocationCount = int

PendingCallback = tuple[TaskExternalId, InvocationCount, BranchId, NodeId]
PendingEventAck = tuple[TaskExternalId, InvocationCount]
PendingEvictionAck = tuple[TaskExternalId, InvocationCount]


class DurableEventListener:
    def __init__(
        self,
        config: ClientConfig,
        admin_client: AdminClient,
        on_server_evict: Callable[[str, int], None] | None = None,
    ):
        self.config = config
        self.token = config.token
        self.admin_client = admin_client

        self._worker_id: str | None = None

        self._conn: grpc.aio.Channel | None = None
        self._stub: V1DispatcherStub | None = None
        self._stream: (
            grpc.aio.StreamStreamCall[DurableTaskRequest, DurableTaskResponse] | None
        ) = None

        self._request_queue: asyncio.Queue[DurableTaskRequest] | None = None
        self._pending_event_acks: dict[
            PendingEventAck, asyncio.Future[DurableTaskEventAck]
        ] = {}
        self._pending_eviction_acks: dict[PendingEvictionAck, asyncio.Future[None]] = {}
        self._pending_callbacks: dict[
            PendingCallback, asyncio.Future[DurableTaskEventLogEntryResult]
        ] = {}

        # Completions that arrived before wait_for_callback() registered a
        # future in _pending_callbacks. This happens when the server delivers
        # an entry_completed between the event ack and the wait_for_callback
        # call (e.g. an already-satisfied sleep delivered via polling).
        self._buffered_completions: TTLCache[
            PendingCallback, DurableTaskEventLogEntryResult
        ] = TTLCache(ttl=timedelta(seconds=10))

        self._receive_task: asyncio.Task[None] | None = None
        self._send_task: asyncio.Task[None] | None = None
        self._running = False
        self._start_lock = asyncio.Lock()

        self._on_server_evict = on_server_evict

    @property
    def worker_id(self) -> str | None:
        return self._worker_id

    async def _connect(self) -> None:
        if self._conn is not None:
            with suppress(Exception):
                await self._conn.close()

        logger.info("durable event listener connecting...")

        self._conn = new_conn(self.config, aio=True)
        self._stub = V1DispatcherStub(self._conn)
        self._request_queue = asyncio.Queue()

        self._stream = cast(
            grpc.aio.StreamStreamCall[DurableTaskRequest, DurableTaskResponse],
            self._stub.DurableTask(
                self._request_iterator(),  # type: ignore[arg-type]
                metadata=get_metadata(self.token),
            ),
        )

        await self._register_worker()
        await self._poll_worker_status()
        logger.info("durable event listener connected")

    async def start(self, worker_id: str) -> None:
        async with self._start_lock:
            if self._running:
                return

            self._worker_id = worker_id
            self._running = True

            await self._connect()

            self._receive_task = asyncio.create_task(self._receive_loop())
            self._send_task = asyncio.create_task(self._send_loop())

    async def ensure_started(self, worker_id: str) -> None:
        if not self._running:
            await self.start(worker_id)

    async def stop(self) -> None:
        self._running = False
        self._buffered_completions.stop_eviction_job()

        self._fail_all_pending(Exception("DurableListener stopped"))

        if self._receive_task:
            self._receive_task.cancel()
            with suppress(asyncio.CancelledError):
                await self._receive_task

        if self._send_task:
            self._send_task.cancel()
            with suppress(asyncio.CancelledError):
                await self._send_task

        if self._conn:
            await self._conn.close()

    async def _request_iterator(self) -> AsyncIterator[DurableTaskRequest]:
        if not self._request_queue:
            raise RuntimeError("Request queue not initialized")

        while self._running:
            with suppress(asyncio.TimeoutError):
                yield await asyncio.wait_for(self._request_queue.get(), timeout=1.0)

    async def _send_loop(self) -> None:
        while self._running:
            await asyncio.sleep(1)
            await self._poll_worker_status()

    async def _poll_worker_status(self) -> None:
        if self._request_queue is None or self._worker_id is None:
            return

        if not self._pending_callbacks:
            return

        waiting = [
            DurableTaskAwaitedCompletedEntry(
                durable_task_external_id=task_ext_id,
                invocation_count=inv_count,
                node_id=node_id,
                branch_id=branch_id,
            )
            for (task_ext_id, inv_count, branch_id, node_id) in self._pending_callbacks
        ]

        request = DurableTaskRequest(
            worker_status=DurableTaskWorkerStatusRequest(
                worker_id=self._worker_id,
                waiting_entries=waiting,
            )
        )
        await self._request_queue.put(request)

    def _fail_pending_acks(self, exc: Exception) -> None:
        for future in self._pending_event_acks.values():
            if not future.done():
                future.set_exception(exc)
        self._pending_event_acks.clear()

        for eviction_future in self._pending_eviction_acks.values():
            if not eviction_future.done():
                eviction_future.set_exception(exc)
        self._pending_eviction_acks.clear()

    def _fail_all_pending(self, exc: Exception) -> None:
        self._fail_pending_acks(exc)

        for future in self._pending_callbacks.values():
            if not future.done():
                future.set_exception(exc)
        self._pending_callbacks.clear()
        self._buffered_completions.clear()

    async def _receive_loop(self) -> None:
        while self._running:
            if not self._stream:
                await asyncio.sleep(DEFAULT_RECONNECT_INTERVAL)
                continue

            try:
                async for response in self._stream:
                    await self._handle_response(response)

                if self._running:
                    logger.warning(
                        f"durable event listener disconnected (EOF), reconnecting in {DEFAULT_RECONNECT_INTERVAL}s..."
                    )
                    self._fail_pending_acks(
                        ConnectionResetError("durable stream disconnected")
                    )
                    await asyncio.sleep(DEFAULT_RECONNECT_INTERVAL)
                    await self._connect()

            except grpc.aio.AioRpcError as e:
                if e.code() == grpc.StatusCode.CANCELLED:
                    break
                logger.warning(
                    f"durable event listener disconnected: code={e.code()}, details={e.details()}, reconnecting in {DEFAULT_RECONNECT_INTERVAL}s..."
                )
                if self._running:
                    self._fail_pending_acks(
                        ConnectionResetError(
                            f"durable stream error: {e.code()} {e.details()}"
                        )
                    )
                    await asyncio.sleep(DEFAULT_RECONNECT_INTERVAL)
                    try:
                        await self._connect()
                    except Exception:
                        logger.exception("failed to reconnect durable event listener")

            except asyncio.CancelledError:
                break

            except Exception as e:
                logger.exception(f"unexpected error in durable event listener: {e}")
                if self._running:
                    self._fail_pending_acks(e)
                    await asyncio.sleep(DEFAULT_RECONNECT_INTERVAL)
                    try:
                        await self._connect()
                    except Exception:
                        logger.exception("failed to reconnect durable event listener")

    async def _handle_response(self, response: DurableTaskResponse) -> None:
        if response.HasField("register_worker"):
            pass
        elif response.HasField("trigger_runs_ack"):
            trigger_ack = response.trigger_runs_ack
            event_key = (
                trigger_ack.durable_task_external_id,
                trigger_ack.invocation_count,
            )
            trigger_ack_future = self._pending_event_acks.pop(event_key, None)
            if trigger_ack_future is not None and not trigger_ack_future.done():
                trigger_ack_future.set_result(
                    DurableTaskEventRunAck(
                        invocation_count=trigger_ack.invocation_count,
                        durable_task_external_id=trigger_ack.durable_task_external_id,
                        run_entries=[
                            DurableTaskRunAckEntry(
                                node_id=e.node_id,
                                branch_id=e.branch_id,
                            )
                            for e in trigger_ack.run_entries
                        ],
                    )
                )
        elif response.HasField("memo_ack"):
            memo_ack = response.memo_ack
            event_key = (
                memo_ack.ref.durable_task_external_id,
                memo_ack.ref.invocation_count,
            )
            memo_ack_future = self._pending_event_acks.pop(event_key, None)
            if memo_ack_future is not None and not memo_ack_future.done():
                memo_ack_future.set_result(
                    DurableTaskEventMemoAck(
                        invocation_count=memo_ack.ref.invocation_count,
                        durable_task_external_id=memo_ack.ref.durable_task_external_id,
                        node_id=memo_ack.ref.node_id,
                        branch_id=memo_ack.ref.branch_id,
                        memo_already_existed=memo_ack.memo_already_existed,
                        memo_result_payload=memo_ack.memo_result_payload,
                    )
                )
        elif response.HasField("wait_for_ack"):
            wait_for_ack = response.wait_for_ack
            event_key = (
                wait_for_ack.ref.durable_task_external_id,
                wait_for_ack.ref.invocation_count,
            )
            wait_for_ack_future = self._pending_event_acks.pop(event_key, None)
            if wait_for_ack_future is not None and not wait_for_ack_future.done():
                wait_for_ack_future.set_result(
                    DurableTaskEventWaitForAck(
                        invocation_count=wait_for_ack.ref.invocation_count,
                        durable_task_external_id=wait_for_ack.ref.durable_task_external_id,
                        node_id=wait_for_ack.ref.node_id,
                        branch_id=wait_for_ack.ref.branch_id,
                    )
                )
        elif response.HasField("entry_completed"):
            completed = response.entry_completed
            completed_key = (
                completed.ref.durable_task_external_id,
                completed.ref.invocation_count,
                completed.ref.branch_id,
                completed.ref.node_id,
            )
            result = DurableTaskEventLogEntryResult.from_proto(completed)
            if completed_key in self._pending_callbacks:
                completed_future = self._pending_callbacks[completed_key]
                if not completed_future.done():
                    completed_future.set_result(result)
                del self._pending_callbacks[completed_key]
            else:
                self._buffered_completions[completed_key] = result
        elif response.HasField("eviction_ack"):
            eviction_ack = response.eviction_ack
            eviction_key = (
                eviction_ack.durable_task_external_id,
                eviction_ack.invocation_count,
            )
            if eviction_key in self._pending_eviction_acks:
                future = self._pending_eviction_acks.pop(eviction_key)
                if not future.done():
                    future.set_result(None)
        elif response.HasField("server_evict"):
            evict = response.server_evict
            logger.info(
                f"received server eviction notification for task {evict.durable_task_external_id} "
                f"invocation {evict.invocation_count}: {evict.reason}"
            )
            self.cleanup_task_state(
                evict.durable_task_external_id, evict.invocation_count
            )
            if self._on_server_evict is not None:
                self._on_server_evict(
                    evict.durable_task_external_id, evict.invocation_count
                )
        elif response.HasField("error"):
            error = response.error
            exc: Exception

            if (
                error.error_type
                == DurableTaskErrorType.DURABLE_TASK_ERROR_TYPE_NONDETERMINISM
            ):
                exc = NonDeterminismError(
                    task_external_id=error.ref.durable_task_external_id,
                    invocation_count=error.ref.invocation_count,
                    message=error.error_message,
                    node_id=error.ref.node_id,
                )
            else:
                ## fallthrough, this shouldn't happen unless we add an error type to the engine and the SDK
                ## hasn't been updated to handle it
                exc = Exception(
                    "Unspecified durable task error: "
                    + error.error_message
                    + f" (type: {error.error_type})"
                )

            event_key = (error.ref.durable_task_external_id, error.ref.invocation_count)
            if event_key in self._pending_event_acks:
                error_pending_ack_future = self._pending_event_acks.pop(event_key)
                if not error_pending_ack_future.done():
                    error_pending_ack_future.set_exception(exc)

            callback_key = (
                error.ref.durable_task_external_id,
                error.ref.invocation_count,
                error.ref.branch_id,
                error.ref.node_id,
            )

            if callback_key in self._pending_callbacks:
                error_pending_callback_future = self._pending_callbacks.pop(
                    callback_key
                )
                if not error_pending_callback_future.done():
                    error_pending_callback_future.set_exception(exc)

            error_eviction_key: PendingEvictionAck = (
                error.ref.durable_task_external_id,
                error.ref.invocation_count,
            )
            if error_eviction_key in self._pending_eviction_acks:
                eviction_future = self._pending_eviction_acks.pop(error_eviction_key)
                if not eviction_future.done():
                    eviction_future.set_exception(exc)

    async def _register_worker(self) -> None:
        if self._request_queue is None or self._worker_id is None:
            raise RuntimeError("Client not started")

        request = DurableTaskRequest(
            register_worker=DurableTaskRequestRegisterWorker(worker_id=self._worker_id)
        )
        await self._request_queue.put(request)

    async def send_event(
        self,
        durable_task_external_id: str,
        invocation_count: int,
        event: DurableTaskSendEvent,
    ) -> DurableTaskEventAck:
        if self._request_queue is None:
            raise RuntimeError("Client not started")

        key = (durable_task_external_id, invocation_count)
        future: asyncio.Future[DurableTaskEventAck] = asyncio.Future()
        self._pending_event_acks[key] = future

        request: DurableTaskRequest

        if isinstance(event, RunChildrenEvent):
            trigger_opts_list = [
                self.admin_client._create_workflow_run_request(
                    workflow_name=child.workflow_name,
                    input=child.input,
                    options=child.trigger_workflow_opts,
                )
                for child in event.children
            ]

            trigger_req = DurableTaskTriggerRunsRequest(
                durable_task_external_id=durable_task_external_id,
                invocation_count=invocation_count,
                trigger_opts=trigger_opts_list,
            )

            request = DurableTaskRequest(trigger_runs=trigger_req)

        elif isinstance(event, WaitForEvent):
            wait_req = DurableTaskWaitForRequest(
                durable_task_external_id=durable_task_external_id,
                invocation_count=invocation_count,
                wait_for_conditions=event.wait_for_conditions,
            )

            request = DurableTaskRequest(wait_for=wait_req)
        elif isinstance(event, MemoEvent):
            memo_req = DurableTaskMemoRequest(
                durable_task_external_id=durable_task_external_id,
                invocation_count=invocation_count,
                key=event.memo_key,
            )

            if event.result is not None:
                memo_req.payload = event.result.encode("utf-8")

            request = DurableTaskRequest(memo=memo_req)

        else:
            e: Never = event
            raise ValueError(f"Unknown durable task send event: {e}")

        await self._request_queue.put(request)

        return await future

    async def wait_for_callback(
        self,
        durable_task_external_id: str,
        invocation_count: int,
        branch_id: int,
        node_id: int,
    ) -> DurableTaskEventLogEntryResult:
        key = (durable_task_external_id, invocation_count, branch_id, node_id)

        if key in self._buffered_completions:
            return self._buffered_completions.pop(key)

        if key not in self._pending_callbacks:
            future: asyncio.Future[DurableTaskEventLogEntryResult] = asyncio.Future()
            self._pending_callbacks[key] = future
            await self._poll_worker_status()

        return await self._pending_callbacks[key]

    def cleanup_task_state(
        self, durable_task_external_id: str, invocation_count: int
    ) -> None:
        """Remove pending callbacks, acks, and buffered completions for old invocations of a task."""
        stale_cb_keys = [
            k
            for k in self._pending_callbacks
            if k[0] == durable_task_external_id and k[1] <= invocation_count
        ]
        for k in stale_cb_keys:
            fut = self._pending_callbacks.pop(k)
            if not fut.done():
                fut.cancel()

        stale_ack_keys = [
            ak
            for ak in self._pending_event_acks
            if ak[0] == durable_task_external_id and ak[1] <= invocation_count
        ]
        for ak in stale_ack_keys:
            ack_fut = self._pending_event_acks.pop(ak)
            if not ack_fut.done():
                ack_fut.cancel()

        stale_early_keys = [
            ek
            for ek in self._buffered_completions
            if ek[0] == durable_task_external_id and ek[1] <= invocation_count
        ]
        for ek in stale_early_keys:
            del self._buffered_completions[ek]

    _EVICTION_ACK_TIMEOUT_S = 30.0

    async def send_evict_invocation(
        self,
        durable_task_external_id: str,
        invocation_count: int,
        reason: str | None = None,
    ) -> None:
        """Send an eviction request to the server and wait for acknowledgement."""
        if self._request_queue is None:
            raise RuntimeError("Client not started")

        eviction_key: PendingEvictionAck = (
            durable_task_external_id,
            invocation_count,
        )
        ack_future: asyncio.Future[None] = asyncio.Future()
        self._pending_eviction_acks[eviction_key] = ack_future

        req = DurableTaskEvictInvocationRequest(
            durable_task_external_id=durable_task_external_id,
            invocation_count=invocation_count,
        )
        if reason is not None:
            req.reason = reason

        request = DurableTaskRequest(evict_invocation=req)
        await self._request_queue.put(request)

        try:
            await asyncio.wait_for(ack_future, timeout=self._EVICTION_ACK_TIMEOUT_S)
        except asyncio.TimeoutError as err:
            self._pending_eviction_acks.pop(eviction_key, None)
            raise TimeoutError(
                f"Eviction ack timed out after {self._EVICTION_ACK_TIMEOUT_S:.0f}s "
                f"for task {durable_task_external_id} invocation {invocation_count}"
            ) from err

    async def send_memo_completed_notification(
        self,
        durable_task_external_id: str,
        node_id: int,
        branch_id: int,
        invocation_count: int,
        memo_key: bytes,
        memo_result_payload: bytes | None,
    ) -> None:
        if self._request_queue is None:
            raise RuntimeError("Client not started")

        await self._request_queue.put(
            DurableTaskRequest(
                complete_memo=DurableTaskCompleteMemoRequest(
                    ref=DurableEventLogEntryRef(
                        durable_task_external_id=durable_task_external_id,
                        node_id=node_id,
                        invocation_count=invocation_count,
                        branch_id=branch_id,
                    ),
                    memo_key=memo_key,
                    payload=memo_result_payload,
                )
            )
        )



from hatchet_sdk.contracts.v1.shared import condition_pb2 as _condition_pb2
from hatchet_sdk.contracts.v1.shared import trigger_pb2 as _trigger_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class DurableTaskErrorType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    DURABLE_TASK_ERROR_TYPE_UNSPECIFIED: _ClassVar[DurableTaskErrorType]
    DURABLE_TASK_ERROR_TYPE_NONDETERMINISM: _ClassVar[DurableTaskErrorType]
DURABLE_TASK_ERROR_TYPE_UNSPECIFIED: DurableTaskErrorType
DURABLE_TASK_ERROR_TYPE_NONDETERMINISM: DurableTaskErrorType

class DurableTaskRequestRegisterWorker(_message.Message):
    __slots__ = ("worker_id",)
    WORKER_ID_FIELD_NUMBER: _ClassVar[int]
    worker_id: str
    def __init__(self, worker_id: _Optional[str] = ...) -> None: ...

class DurableTaskResponseRegisterWorker(_message.Message):
    __slots__ = ("worker_id",)
    WORKER_ID_FIELD_NUMBER: _ClassVar[int]
    worker_id: str
    def __init__(self, worker_id: _Optional[str] = ...) -> None: ...

class DurableEventLogEntryRef(_message.Message):
    __slots__ = ("durable_task_external_id", "invocation_count", "branch_id", "node_id")
    DURABLE_TASK_EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    INVOCATION_COUNT_FIELD_NUMBER: _ClassVar[int]
    BRANCH_ID_FIELD_NUMBER: _ClassVar[int]
    NODE_ID_FIELD_NUMBER: _ClassVar[int]
    durable_task_external_id: str
    invocation_count: int
    branch_id: int
    node_id: int
    def __init__(self, durable_task_external_id: _Optional[str] = ..., invocation_count: _Optional[int] = ..., branch_id: _Optional[int] = ..., node_id: _Optional[int] = ...) -> None: ...

class DurableTaskRunAckEntry(_message.Message):
    __slots__ = ("node_id", "branch_id")
    NODE_ID_FIELD_NUMBER: _ClassVar[int]
    BRANCH_ID_FIELD_NUMBER: _ClassVar[int]
    node_id: int
    branch_id: int
    def __init__(self, node_id: _Optional[int] = ..., branch_id: _Optional[int] = ...) -> None: ...

class DurableTaskEventMemoAckResponse(_message.Message):
    __slots__ = ("ref", "memo_already_existed", "memo_result_payload")
    REF_FIELD_NUMBER: _ClassVar[int]
    MEMO_ALREADY_EXISTED_FIELD_NUMBER: _ClassVar[int]
    MEMO_RESULT_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    ref: DurableEventLogEntryRef
    memo_already_existed: bool
    memo_result_payload: bytes
    def __init__(self, ref: _Optional[_Union[DurableEventLogEntryRef, _Mapping]] = ..., memo_already_existed: bool = ..., memo_result_payload: _Optional[bytes] = ...) -> None: ...

class DurableTaskEventTriggerRunsAckResponse(_message.Message):
    __slots__ = ("durable_task_external_id", "invocation_count", "run_entries")
    DURABLE_TASK_EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    INVOCATION_COUNT_FIELD_NUMBER: _ClassVar[int]
    RUN_ENTRIES_FIELD_NUMBER: _ClassVar[int]
    durable_task_external_id: str
    invocation_count: int
    run_entries: _containers.RepeatedCompositeFieldContainer[DurableTaskRunAckEntry]
    def __init__(self, durable_task_external_id: _Optional[str] = ..., invocation_count: _Optional[int] = ..., run_entries: _Optional[_Iterable[_Union[DurableTaskRunAckEntry, _Mapping]]] = ...) -> None: ...

class DurableTaskEventWaitForAckResponse(_message.Message):
    __slots__ = ("ref",)
    REF_FIELD_NUMBER: _ClassVar[int]
    ref: DurableEventLogEntryRef
    def __init__(self, ref: _Optional[_Union[DurableEventLogEntryRef, _Mapping]] = ...) -> None: ...

class DurableTaskEventLogEntryCompletedResponse(_message.Message):
    __slots__ = ("ref", "payload")
    REF_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    ref: DurableEventLogEntryRef
    payload: bytes
    def __init__(self, ref: _Optional[_Union[DurableEventLogEntryRef, _Mapping]] = ..., payload: _Optional[bytes] = ...) -> None: ...

class DurableTaskEvictInvocationRequest(_message.Message):
    __slots__ = ("invocation_count", "durable_task_external_id", "reason")
    INVOCATION_COUNT_FIELD_NUMBER: _ClassVar[int]
    DURABLE_TASK_EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    invocation_count: int
    durable_task_external_id: str
    reason: str
    def __init__(self, invocation_count: _Optional[int] = ..., durable_task_external_id: _Optional[str] = ..., reason: _Optional[str] = ...) -> None: ...

class DurableTaskEvictionAckResponse(_message.Message):
    __slots__ = ("invocation_count", "durable_task_external_id")
    INVOCATION_COUNT_FIELD_NUMBER: _ClassVar[int]
    DURABLE_TASK_EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    invocation_count: int
    durable_task_external_id: str
    def __init__(self, invocation_count: _Optional[int] = ..., durable_task_external_id: _Optional[str] = ...) -> None: ...

class DurableTaskAwaitedCompletedEntry(_message.Message):
    __slots__ = ("durable_task_external_id", "branch_id", "node_id", "invocation_count")
    DURABLE_TASK_EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    BRANCH_ID_FIELD_NUMBER: _ClassVar[int]
    NODE_ID_FIELD_NUMBER: _ClassVar[int]
    INVOCATION_COUNT_FIELD_NUMBER: _ClassVar[int]
    durable_task_external_id: str
    branch_id: int
    node_id: int
    invocation_count: int
    def __init__(self, durable_task_external_id: _Optional[str] = ..., branch_id: _Optional[int] = ..., node_id: _Optional[int] = ..., invocation_count: _Optional[int] = ...) -> None: ...

class DurableTaskServerEvictNotice(_message.Message):
    __slots__ = ("durable_task_external_id", "invocation_count", "reason")
    DURABLE_TASK_EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    INVOCATION_COUNT_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    durable_task_external_id: str
    invocation_count: int
    reason: str
    def __init__(self, durable_task_external_id: _Optional[str] = ..., invocation_count: _Optional[int] = ..., reason: _Optional[str] = ...) -> None: ...

class DurableTaskWorkerStatusRequest(_message.Message):
    __slots__ = ("worker_id", "waiting_entries")
    WORKER_ID_FIELD_NUMBER: _ClassVar[int]
    WAITING_ENTRIES_FIELD_NUMBER: _ClassVar[int]
    worker_id: str
    waiting_entries: _containers.RepeatedCompositeFieldContainer[DurableTaskAwaitedCompletedEntry]
    def __init__(self, worker_id: _Optional[str] = ..., waiting_entries: _Optional[_Iterable[_Union[DurableTaskAwaitedCompletedEntry, _Mapping]]] = ...) -> None: ...

class DurableTaskCompleteMemoRequest(_message.Message):
    __slots__ = ("ref", "payload", "memo_key")
    REF_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    MEMO_KEY_FIELD_NUMBER: _ClassVar[int]
    ref: DurableEventLogEntryRef
    payload: bytes
    memo_key: bytes
    def __init__(self, ref: _Optional[_Union[DurableEventLogEntryRef, _Mapping]] = ..., payload: _Optional[bytes] = ..., memo_key: _Optional[bytes] = ...) -> None: ...

class DurableTaskMemoRequest(_message.Message):
    __slots__ = ("invocation_count", "durable_task_external_id", "key", "payload")
    INVOCATION_COUNT_FIELD_NUMBER: _ClassVar[int]
    DURABLE_TASK_EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    KEY_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    invocation_count: int
    durable_task_external_id: str
    key: bytes
    payload: bytes
    def __init__(self, invocation_count: _Optional[int] = ..., durable_task_external_id: _Optional[str] = ..., key: _Optional[bytes] = ..., payload: _Optional[bytes] = ...) -> None: ...

class DurableTaskTriggerRunsRequest(_message.Message):
    __slots__ = ("invocation_count", "durable_task_external_id", "trigger_opts")
    INVOCATION_COUNT_FIELD_NUMBER: _ClassVar[int]
    DURABLE_TASK_EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    TRIGGER_OPTS_FIELD_NUMBER: _ClassVar[int]
    invocation_count: int
    durable_task_external_id: str
    trigger_opts: _containers.RepeatedCompositeFieldContainer[_trigger_pb2.TriggerWorkflowRequest]
    def __init__(self, invocation_count: _Optional[int] = ..., durable_task_external_id: _Optional[str] = ..., trigger_opts: _Optional[_Iterable[_Union[_trigger_pb2.TriggerWorkflowRequest, _Mapping]]] = ...) -> None: ...

class DurableTaskWaitForRequest(_message.Message):
    __slots__ = ("invocation_count", "durable_task_external_id", "wait_for_conditions")
    INVOCATION_COUNT_FIELD_NUMBER: _ClassVar[int]
    DURABLE_TASK_EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    WAIT_FOR_CONDITIONS_FIELD_NUMBER: _ClassVar[int]
    invocation_count: int
    durable_task_external_id: str
    wait_for_conditions: _condition_pb2.DurableEventListenerConditions
    def __init__(self, invocation_count: _Optional[int] = ..., durable_task_external_id: _Optional[str] = ..., wait_for_conditions: _Optional[_Union[_condition_pb2.DurableEventListenerConditions, _Mapping]] = ...) -> None: ...

class DurableTaskRequest(_message.Message):
    __slots__ = ("register_worker", "memo", "trigger_runs", "wait_for", "evict_invocation", "worker_status", "complete_memo")
    REGISTER_WORKER_FIELD_NUMBER: _ClassVar[int]
    MEMO_FIELD_NUMBER: _ClassVar[int]
    TRIGGER_RUNS_FIELD_NUMBER: _ClassVar[int]
    WAIT_FOR_FIELD_NUMBER: _ClassVar[int]
    EVICT_INVOCATION_FIELD_NUMBER: _ClassVar[int]
    WORKER_STATUS_FIELD_NUMBER: _ClassVar[int]
    COMPLETE_MEMO_FIELD_NUMBER: _ClassVar[int]
    register_worker: DurableTaskRequestRegisterWorker
    memo: DurableTaskMemoRequest
    trigger_runs: DurableTaskTriggerRunsRequest
    wait_for: DurableTaskWaitForRequest
    evict_invocation: DurableTaskEvictInvocationRequest
    worker_status: DurableTaskWorkerStatusRequest
    complete_memo: DurableTaskCompleteMemoRequest
    def __init__(self, register_worker: _Optional[_Union[DurableTaskRequestRegisterWorker, _Mapping]] = ..., memo: _Optional[_Union[DurableTaskMemoRequest, _Mapping]] = ..., trigger_runs: _Optional[_Union[DurableTaskTriggerRunsRequest, _Mapping]] = ..., wait_for: _Optional[_Union[DurableTaskWaitForRequest, _Mapping]] = ..., evict_invocation: _Optional[_Union[DurableTaskEvictInvocationRequest, _Mapping]] = ..., worker_status: _Optional[_Union[DurableTaskWorkerStatusRequest, _Mapping]] = ..., complete_memo: _Optional[_Union[DurableTaskCompleteMemoRequest, _Mapping]] = ...) -> None: ...

class DurableTaskErrorResponse(_message.Message):
    __slots__ = ("ref", "error_type", "error_message")
    REF_FIELD_NUMBER: _ClassVar[int]
    ERROR_TYPE_FIELD_NUMBER: _ClassVar[int]
    ERROR_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    ref: DurableEventLogEntryRef
    error_type: DurableTaskErrorType
    error_message: str
    def __init__(self, ref: _Optional[_Union[DurableEventLogEntryRef, _Mapping]] = ..., error_type: _Optional[_Union[DurableTaskErrorType, str]] = ..., error_message: _Optional[str] = ...) -> None: ...

class DurableTaskResponse(_message.Message):
    __slots__ = ("register_worker", "memo_ack", "trigger_runs_ack", "wait_for_ack", "entry_completed", "error", "eviction_ack", "server_evict")
    REGISTER_WORKER_FIELD_NUMBER: _ClassVar[int]
    MEMO_ACK_FIELD_NUMBER: _ClassVar[int]
    TRIGGER_RUNS_ACK_FIELD_NUMBER: _ClassVar[int]
    WAIT_FOR_ACK_FIELD_NUMBER: _ClassVar[int]
    ENTRY_COMPLETED_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    EVICTION_ACK_FIELD_NUMBER: _ClassVar[int]
    SERVER_EVICT_FIELD_NUMBER: _ClassVar[int]
    register_worker: DurableTaskResponseRegisterWorker
    memo_ack: DurableTaskEventMemoAckResponse
    trigger_runs_ack: DurableTaskEventTriggerRunsAckResponse
    wait_for_ack: DurableTaskEventWaitForAckResponse
    entry_completed: DurableTaskEventLogEntryCompletedResponse
    error: DurableTaskErrorResponse
    eviction_ack: DurableTaskEvictionAckResponse
    server_evict: DurableTaskServerEvictNotice
    def __init__(self, register_worker: _Optional[_Union[DurableTaskResponseRegisterWorker, _Mapping]] = ..., memo_ack: _Optional[_Union[DurableTaskEventMemoAckResponse, _Mapping]] = ..., trigger_runs_ack: _Optional[_Union[DurableTaskEventTriggerRunsAckResponse, _Mapping]] = ..., wait_for_ack: _Optional[_Union[DurableTaskEventWaitForAckResponse, _Mapping]] = ..., entry_completed: _Optional[_Union[DurableTaskEventLogEntryCompletedResponse, _Mapping]] = ..., error: _Optional[_Union[DurableTaskErrorResponse, _Mapping]] = ..., eviction_ack: _Optional[_Union[DurableTaskEvictionAckResponse, _Mapping]] = ..., server_evict: _Optional[_Union[DurableTaskServerEvictNotice, _Mapping]] = ...) -> None: ...

class RegisterDurableEventRequest(_message.Message):
    __slots__ = ("task_id", "signal_key", "conditions")
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    SIGNAL_KEY_FIELD_NUMBER: _ClassVar[int]
    CONDITIONS_FIELD_NUMBER: _ClassVar[int]
    task_id: str
    signal_key: str
    conditions: _condition_pb2.DurableEventListenerConditions
    def __init__(self, task_id: _Optional[str] = ..., signal_key: _Optional[str] = ..., conditions: _Optional[_Union[_condition_pb2.DurableEventListenerConditions, _Mapping]] = ...) -> None: ...

class RegisterDurableEventResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ListenForDurableEventRequest(_message.Message):
    __slots__ = ("task_id", "signal_key")
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    SIGNAL_KEY_FIELD_NUMBER: _ClassVar[int]
    task_id: str
    signal_key: str
    def __init__(self, task_id: _Optional[str] = ..., signal_key: _Optional[str] = ...) -> None: ...

class DurableEvent(_message.Message):
    __slots__ = ("task_id", "signal_key", "data")
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    SIGNAL_KEY_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    task_id: str
    signal_key: str
    data: bytes
    def __init__(self, task_id: _Optional[str] = ..., signal_key: _Optional[str] = ..., data: _Optional[bytes] = ...) -> None: ...



from __future__ import annotations

import asyncio
import hashlib
import json
from collections.abc import Awaitable, Callable
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Any, ParamSpec, TypeVar, cast, overload
from warnings import warn

from pydantic import BaseModel, TypeAdapter

from hatchet_sdk.clients.admin import (
    AdminClient,
    WorkflowRunTriggerConfig,
)
from hatchet_sdk.clients.dispatcher.dispatcher import (  # type: ignore[attr-defined]
    Action,
    DispatcherClient,
)
from hatchet_sdk.clients.events import EventClient
from hatchet_sdk.clients.listeners.durable_event_listener import (
    DurableEventListener,
    DurableTaskEventMemoAck,
    DurableTaskEventRunAck,
    DurableTaskEventWaitForAck,
    MemoEvent,
    RunChildEvent,
    RunChildrenEvent,
    WaitForEvent,
)
from hatchet_sdk.clients.listeners.legacy.pre_eviction_durable_event_listener import (
    PreEvictionDurableEventListener,
)
from hatchet_sdk.conditions import (
    OrGroup,
    SleepCondition,
    UserEventCondition,
    build_conditions_proto,
    flatten_conditions,
)
from hatchet_sdk.context.pre_eviction import aio_wait_for_pre_eviction
from hatchet_sdk.context.worker_context import WorkerContext
from hatchet_sdk.deprecated.deprecation import semver_less_than
from hatchet_sdk.engine_version import MinEngineVersion
from hatchet_sdk.exceptions import TaskRunError
from hatchet_sdk.features.runs import RunsClient
from hatchet_sdk.logger import logger
from hatchet_sdk.runnables.types import (
    R,
    TWorkflowInput,
    ValidTaskReturnType,
)
from hatchet_sdk.serde import HATCHET_PYDANTIC_SENTINEL
from hatchet_sdk.utils.timedelta_to_expression import (
    Duration,
    expr_to_timedelta,
    timedelta_to_expr,
)
from hatchet_sdk.utils.typing import (
    DataclassInstance,
    JSONSerializableMapping,
    LogLevel,
)
from hatchet_sdk.worker.durable_eviction.instrumentation import (
    aio_durable_eviction_wait,
)
from hatchet_sdk.worker.durable_eviction.manager import DurableEvictionManager
from hatchet_sdk.worker.runner.utils.capture_logs import AsyncLogSender, LogRecord

PMemo = ParamSpec("PMemo")
TMemo = TypeVar("TMemo", bound=ValidTaskReturnType)

if TYPE_CHECKING:
    from hatchet_sdk.runnables.task import Task


TPayload = TypeVar("TPayload", bound=BaseModel | DataclassInstance | dict[str, Any])


class SleepResult(BaseModel):
    duration: timedelta


class MemoNowResult(BaseModel):
    ts: datetime


def _compute_memo_key(task_run_external_id: str, *args: Any, **kwargs: Any) -> bytes:
    h = hashlib.sha256()
    h.update(task_run_external_id.encode())
    h.update(json.dumps(args, default=str, sort_keys=True).encode())
    h.update(json.dumps(kwargs, default=str, sort_keys=True).encode())
    return h.digest()


class Context:
    def __init__(
        self,
        action: Action,
        dispatcher_client: DispatcherClient,
        admin_client: AdminClient,
        event_client: EventClient,
        durable_event_listener: (
            DurableEventListener | PreEvictionDurableEventListener | None
        ),
        worker: WorkerContext,
        runs_client: RunsClient,
        lifespan_context: Any | None,
        log_sender: AsyncLogSender,
        max_attempts: int,
        task_name: str,
        workflow_name: str,
    ):
        self.worker = worker

        self.data = action.action_payload

        self.action = action

        self.step_run_id = action.step_run_id
        self.exit_flag = False
        self.dispatcher_client = dispatcher_client
        self.admin_client = admin_client
        self.event_client = event_client
        self.runs_client = runs_client
        self.durable_event_listener = durable_event_listener

        self.input = self.data.input
        self.filter_payload = self.data.filter_payload
        self.log_sender = log_sender

        self._lifespan_context = lifespan_context

        self.stream_index = 0
        self._max_attempts = max_attempts
        self._workflow_name = workflow_name
        self._task_name = task_name

    def _increment_stream_index(self) -> int:
        index = self.stream_index
        self.stream_index += 1

        return index

    def was_skipped(self, task: Task[TWorkflowInput, R]) -> bool:
        """
        Check if a given task was skipped. You can read about skipping in [the docs](https://docs.hatchet.run/home/conditional-workflows#skip_if).

        :param task: The task to check the status of (skipped or not).
        :return: True if the task was skipped, False otherwise.
        """
        return self.data.parents.get(task.name, {}).get("skipped", False) is True

    @property
    def trigger_data(self) -> JSONSerializableMapping:
        return self.data.triggers

    def task_output(self, task: Task[TWorkflowInput, R]) -> R:
        """
        Get the output of a parent task in a DAG.

        :param task: The task whose output you want to retrieve.
        :return: The output of the parent task, validated against the task's validators.
        :raises ValueError: If the task was skipped or if the step output for the task is not found.
        """
        from hatchet_sdk.runnables.types import R
        from hatchet_sdk.serde import HATCHET_PYDANTIC_SENTINEL

        if self.was_skipped(task):
            raise ValueError(f"{task.name} was skipped")

        try:
            parent_step_data = cast(R, self.data.parents[task.name])
        except KeyError as e:
            raise ValueError(f"Step output for '{task.name}' not found") from e

        return cast(
            R,
            task.validators.step_output.validate_python(
                parent_step_data, context=HATCHET_PYDANTIC_SENTINEL
            ),
        )

    def aio_task_output(self, task: Task[TWorkflowInput, R]) -> R:
        warn(
            "`aio_task_output` is deprecated. Use `task_output` instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        if task.is_async_function:
            return self.task_output(task)

        raise ValueError(
            f"Task '{task.name}' is not an async function. Use `task_output` instead."
        )

    @property
    def was_triggered_by_event(self) -> bool:
        """
        A property that indicates whether the workflow was triggered by an event.

        :return: True if the workflow was triggered by an event, False otherwise.
        """
        return self.data.triggered_by == "event"

    @property
    def workflow_input(self) -> JSONSerializableMapping:
        """
        The input to the workflow, as a dictionary. It's recommended to use the `input` parameter to the task (the first argument passed into the task at runtime) instead of this property.

        :return: The input to the workflow.
        """
        return self.input

    @property
    def lifespan(self) -> Any:
        """
        The worker lifespan, if it exists. You can read about lifespans in [the docs](https://docs.hatchet.run/home/lifespans).

        **Note: You'll need to cast the return type of this property to the type returned by your lifespan generator.**
        """
        return self._lifespan_context

    @property
    def workflow_run_id(self) -> str:
        """
        The id of the current workflow run.

        :return: The id of the current workflow run.
        """
        return self.action.workflow_run_id

    def _set_cancellation_flag(self) -> None:
        self.exit_flag = True

    def cancel(self) -> None:
        """
        Cancel the current task run. This will call the Hatchet API to cancel the step run and set the exit flag to True.

        :return: None
        """
        logger.debug("cancelling step...")
        self.runs_client.cancel(self.step_run_id)
        self._set_cancellation_flag()

    async def aio_cancel(self) -> None:
        """
        Cancel the current task run. This will call the Hatchet API to cancel the step run and set the exit flag to True.

        :return: None
        """
        logger.debug("cancelling step...")
        await self.runs_client.aio_cancel(self.step_run_id)
        self._set_cancellation_flag()

    def done(self) -> bool:
        """
        Check if the current task run has been cancelled.

        :return: True if the task run has been cancelled, False otherwise.
        """
        return self.exit_flag

    def log(
        self, line: str | JSONSerializableMapping, raise_on_error: bool = False
    ) -> None:
        """
        Log a line to the Hatchet API. This will send the log line to the Hatchet API and return immediately.

        :param line: The line to log. Can be a string or a JSON serializable mapping.
        :param raise_on_error: If True, will raise an exception if the log fails. Defaults to False.
        :return: None
        """

        if self.step_run_id == "":
            return

        if not isinstance(line, str):
            try:
                line = json.dumps(line)
            except Exception:
                line = str(line)

        logger.info(line)
        self.log_sender.publish(
            LogRecord(
                message=line,
                step_run_id=self.step_run_id,
                level=LogLevel.INFO,
                task_retry_count=self.retry_count,
            )
        )

    def release_slot(self) -> None:
        """
        Manually release the slot for the current step run to free up a slot on the worker. Note that this is an advanced feature and should be used with caution.

        :return: None
        """
        return self.dispatcher_client.release_slot(self.step_run_id)

    def put_stream(self, data: str | bytes) -> None:
        """
        Put a stream event to the Hatchet API. This will send the data to the Hatchet API and return immediately. You can then subscribe to the stream from a separate consumer.

        :param data: The data to send to the Hatchet API. Can be a string or bytes.
        :return: None
        """
        try:
            ix = self._increment_stream_index()

            self.event_client.stream(
                data=data,
                step_run_id=self.step_run_id,
                index=ix,
            )
        except Exception:
            logger.exception("error putting stream event")

    async def aio_put_stream(self, data: str | bytes) -> None:
        """
        Put a stream event to the Hatchet API. This will send the data to the Hatchet API and return immediately. You can then subscribe to the stream from a separate consumer.

        :param data: The data to send to the Hatchet API. Can be a string or bytes.
        :return: None
        """
        await asyncio.to_thread(self.put_stream, data)

    def refresh_timeout(self, increment_by: str | timedelta) -> None:
        """
        Refresh the timeout for the current task run. You can read about refreshing timeouts in [the docs](https://docs.hatchet.run/home/timeouts#refreshing-timeouts).

        :param increment_by: The amount of time to increment the timeout by. Can be a string (e.g. "5m") or a timedelta object.
        :return: None
        """
        if isinstance(increment_by, timedelta):
            increment_by = timedelta_to_expr(increment_by)

        try:
            return self.dispatcher_client.refresh_timeout(
                step_run_id=self.step_run_id, increment_by=increment_by
            )
        except Exception:
            logger.exception("error refreshing timeout")

    @property
    def retry_count(self) -> int:
        """
        The retry count of the current task run, which corresponds to the number of times the task has been retried.

        :return: The retry count of the current task run.
        """
        return self.action.retry_count

    @property
    def attempt_number(self) -> int:
        """
        The attempt number of the current task run, which corresponds to the number of times the task has been attempted, including the initial attempt. This is one more than the retry count.

        :return: The attempt number of the current task run.
        """

        return self.retry_count + 1

    @property
    def max_attempts(self) -> int:
        """
        The maximum number of attempts allowed for the current task run, computed as the number of retries plus one.

        :return: The maximum number of attempts allowed for the current task run.
        """

        return self._max_attempts

    @property
    def additional_metadata(self) -> JSONSerializableMapping:
        """
        The additional metadata sent with the current task run.

        :return: The additional metadata sent with the current task run, or None if no additional metadata was sent.
        """
        return self.action.additional_metadata

    @property
    def child_index(self) -> int | None:
        return self.action.child_workflow_index

    @property
    def child_key(self) -> str | None:
        return self.action.child_workflow_key

    @property
    def parent_workflow_run_id(self) -> str | None:
        """
        The parent workflow run id of the current task run, if it exists. This is useful for knowing which workflow run spawned this run as a child.

        :return: The parent workflow run id of the current task run, or None if it does not exist.
        """
        return self.action.parent_workflow_run_id

    @property
    def priority(self) -> int | None:
        """
        The priority that the current task was run with.

        :return: The priority of the current task run, or None if no priority was set.
        """
        return self.action.priority

    @property
    def workflow_id(self) -> str | None:
        """
        The id of the workflow that this task belongs to.

        :return: The id of the workflow that this task belongs to.
        """

        return self.action.workflow_id

    @property
    def workflow_version_id(self) -> str | None:
        """
        The id of the workflow version that this task belongs to.

        :return: The id of the workflow version that this task belongs to.
        """

        return self.action.workflow_version_id

    @property
    def task_run_errors(self) -> dict[str, str]:
        """
        A helper intended to be used in an on-failure step to retrieve the errors that occurred in upstream task runs.

        :return: A dictionary mapping task names to their error messages.
        """
        errors = self.data.step_run_errors

        if not errors:
            logger.error(
                "no step run errors found. `context.task_run_errors` is intended to be run in an on-failure step, and will only work on engine versions more recent than v0.53.10"
            )

        return errors

    @property
    def workflow_name(self) -> str:
        return self._workflow_name

    @property
    def task_name(self) -> str:
        return self._task_name

    @property
    def worker_id(self) -> str:
        return self.action.worker_id

    def fetch_task_run_error(
        self,
        task: Task[TWorkflowInput, R],
    ) -> str | None:
        """
        **DEPRECATED**: Use `get_task_run_error` instead.

        A helper intended to be used in an on-failure step to retrieve the error that occurred in a specific upstream task run.

        :param task: The task whose error you want to retrieve.
        :return: The error message of the task run, or None if no error occurred.
        """
        warn(
            "`fetch_task_run_error` is deprecated. Use `get_task_run_error` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        errors = self.data.step_run_errors

        return errors.get(task.name)

    def get_task_run_error(
        self,
        task: Task[TWorkflowInput, R],
    ) -> TaskRunError | None:
        """
        A helper intended to be used in an on-failure step to retrieve the error that occurred in a specific upstream task run.

        :param task: The task whose error you want to retrieve.
        :return: The error message of the task run, or None if no error occurred.
        """
        errors = self.data.step_run_errors

        error = errors.get(task.name)

        if not error:
            return None

        return TaskRunError.deserialize(error)


class DurableContext(Context):
    def __init__(
        self,
        action: Action,
        dispatcher_client: DispatcherClient,
        admin_client: AdminClient,
        event_client: EventClient,
        durable_event_listener: (
            DurableEventListener | PreEvictionDurableEventListener | None
        ),
        worker: WorkerContext,
        runs_client: RunsClient,
        lifespan_context: Any | None,
        log_sender: AsyncLogSender,
        max_attempts: int,
        task_name: str,
        workflow_name: str,
        durable_eviction_manager: DurableEvictionManager | None = None,
        engine_version: str | None = None,
    ):
        super().__init__(
            action,
            dispatcher_client,
            admin_client,
            event_client,
            durable_event_listener,
            worker,
            runs_client,
            lifespan_context,
            log_sender,
            max_attempts,
            task_name,
            workflow_name,
        )

        self._wait_index = 0
        self._durable_eviction_manager = durable_eviction_manager
        self._engine_version = engine_version

    @property
    def _durable_listener(self) -> DurableEventListener:
        if self.durable_event_listener is None:
            raise ValueError("Durable task client is not available")

        if not isinstance(self.durable_event_listener, DurableEventListener):
            raise TypeError(
                "Expected DurableEventListener, got "
                f"{type(self.durable_event_listener).__name__}"
            )
        return self.durable_event_listener

    @property
    def _supports_durable_eviction(self) -> bool:
        if not self._engine_version:
            return False
        return not semver_less_than(
            self._engine_version, MinEngineVersion.DURABLE_EVICTION
        )

    @property
    def wait_index(self) -> int:
        return self._wait_index

    def _increment_wait_index(self) -> int:
        index = self._wait_index
        self._wait_index += 1

        return index

    ## IMPORTANT: This method is instrumented by HatchetInstrumentor._wrap_aio_wait_for.
    ## Keep the signature in sync with the instrumentor wrapper.
    async def aio_wait_for(
        self,
        signal_key: str,
        *conditions: SleepCondition | UserEventCondition | OrGroup,
    ) -> dict[str, Any]:
        """
        Durably wait for either a sleep or an event.

        :param signal_key: The key to use for the durable event. This is used to identify the event in the Hatchet API.
        :param *conditions: The conditions to wait for. Can be a SleepCondition or UserEventCondition.

        :return: A dictionary containing the results of the wait.

        :raises ValueError: If the durable task client is not available.
        :raises TypeError: If the durable event listener is not of type DurableEventListener or PreEvictionDurableEventListener.
        """
        if self.durable_event_listener is None:
            raise ValueError("Durable task client is not available")

        if not self._supports_durable_eviction:
            return await aio_wait_for_pre_eviction(self, signal_key, *conditions)

        listener = self._durable_listener

        await self._ensure_stream_started()

        flat_conditions = flatten_conditions(list(conditions))
        conditions_proto = build_conditions_proto(
            flat_conditions, self.runs_client.client_config
        )
        ack = await listener.send_event(
            durable_task_external_id=self.step_run_id,
            invocation_count=self.invocation_count,
            event=WaitForEvent(wait_for_conditions=conditions_proto),
        )

        if not isinstance(ack, DurableTaskEventWaitForAck):
            raise TypeError(f"Expected wait-for ack, got {type(ack).__name__}")

        node_id = ack.node_id
        branch_id = ack.branch_id

        async with aio_durable_eviction_wait(
            wait_kind="wait_for",
            resource_id=signal_key,
            action_key=self.action.key,
            eviction_manager=self._durable_eviction_manager,
        ):
            result = await listener.wait_for_callback(
                durable_task_external_id=self.step_run_id,
                node_id=node_id,
                branch_id=branch_id,
                invocation_count=self.invocation_count,
            )

        return result.payload or {}

    async def aio_sleep_for(self, duration: Duration) -> SleepResult:
        """
        Lightweight wrapper for durable sleep. Allows for shorthand usage of `ctx.aio_wait_for` when specifying a sleep condition.

        For more complicated conditions, use `ctx.aio_wait_for` directly.
        """

        wait_index = self._increment_wait_index()

        res = await self.aio_wait_for(
            f"sleep:{timedelta_to_expr(duration)}-{wait_index}",
            SleepCondition(duration=duration),
        )

        ## lots of implicit use of engine semantics / internal logic here.
        ## the engine returns an object like this:
        ## {"CREATE": {"signal_key_1": [{"id": ...}]}}
        ## since we have a single match we're looking for, we know that
        ## the list of matches will only have one item, so we can extract and parse it
        matches: dict[str, list[dict[str, Any]]] = res.get("CREATE", {})
        _, raw_matches = next(iter(matches.items()))
        sleep = raw_matches[0]

        return SleepResult(
            duration=expr_to_timedelta(
                sleep.get("sleep_duration", timedelta_to_expr(duration))
            )
        )

    @overload
    async def aio_wait_for_event(
        self,
        key: str,
        expression: str | None = None,
        *,
        payload_validator: type[TPayload],
    ) -> TPayload: ...

    @overload
    async def aio_wait_for_event(
        self,
        key: str,
        expression: str | None = None,
    ) -> dict[str, Any]: ...

    async def aio_wait_for_event(
        self,
        key: str,
        expression: str | None = None,
        *,
        payload_validator: type[Any] | None = None,
    ) -> Any:
        """
        Lightweight wrapper for waiting for a user event. Allows for shorthand usage of `ctx.aio_wait_for` when specifying a user event condition.

        For more complicated conditions, use `ctx.aio_wait_for` directly.

        :param key: The event key to wait for.
        :param expression: An optional CEL expression to filter events.
        :param payload_validator: An optional type (e.g. a Pydantic model, dataclass, or TypedDict) to validate the event payload against. If provided, the payload will be validated and returned as an instance of this type.

        :return: The payload of the event, validated against the provided payload_validator if it was given, or as a raw dictionary if no payload_validator was provided.
        """

        wait_index = self._increment_wait_index()

        result = await self.aio_wait_for(
            f"event:{key}-{wait_index}",
            UserEventCondition(event_key=key, expression=expression),
        )

        ## lots of implicit use of engine semantics / internal logic here.
        ## the engine returns an object like this:
        ## {"CREATE": {"signal_key_1": [{"id": ...}]}}
        ## since we have a single match we're looking for, we know that
        ## the list of matches will only have one item, so we can extract and parse it
        matches: dict[str, list[dict[str, Any]]] = result.get("CREATE", {})
        _, raw_matches = next(iter(matches.items()))
        raw_payload = raw_matches[0]

        if payload_validator is not None:
            adapter = TypeAdapter(payload_validator)
            return adapter.validate_python(
                raw_payload, context=HATCHET_PYDANTIC_SENTINEL
            )

        return raw_payload

    ## IMPORTANT: This method is instrumented by HatchetInstrumentor._wrap_spawn_children_no_wait.
    ## Keep the signature in sync with the instrumentor wrapper.
    async def _spawn_children_no_wait(
        self,
        configs: list[WorkflowRunTriggerConfig],
    ) -> list[tuple[int, int, str]]:
        listener = self._durable_listener

        await self._ensure_stream_started()

        ack = await listener.send_event(
            durable_task_external_id=self.step_run_id,
            invocation_count=self.invocation_count,
            event=RunChildrenEvent(
                children=[
                    RunChildEvent(
                        workflow_name=c.workflow_name,
                        input=c.input,
                        trigger_workflow_opts=c.options,
                    )
                    for c in configs
                ]
            ),
        )

        if not isinstance(ack, DurableTaskEventRunAck):
            raise TypeError(f"Expected run ack, got {type(ack).__name__}")

        return [
            (entry.node_id, entry.branch_id, configs[i].workflow_name)
            for i, entry in enumerate(ack.run_entries)
        ]

    async def _aio_result_for_spawned_child(
        self,
        node_id: int,
        branch_id: int,
        workflow_name: str,
    ) -> dict[str, Any]:
        listener = self._durable_listener

        async with aio_durable_eviction_wait(
            wait_kind="spawn_child",
            resource_id=workflow_name,
            action_key=self.action.key,
            eviction_manager=self._durable_eviction_manager,
        ):
            result = await listener.wait_for_callback(
                durable_task_external_id=self.step_run_id,
                node_id=node_id,
                branch_id=branch_id,
                invocation_count=self.invocation_count,
            )

        return result.payload or {}

    async def _ensure_stream_started(self) -> None:
        if not isinstance(self.durable_event_listener, DurableEventListener):
            raise ValueError("Durable task client is not available")

        await self.durable_event_listener.ensure_started(self.action.worker_id)

    @property
    def invocation_count(self) -> int:
        return self.action.durable_task_invocation_count or 1

    ## IMPORTANT: This method is instrumented by HatchetInstrumentor._wrap_aio_memo.
    ## Keep the signature in sync with the instrumentor wrapper.
    async def _aio_memo(
        self,
        fn: Callable[PMemo, Awaitable[TMemo]],
        result_validator: type[TMemo],
        /,
        *args: PMemo.args,
        **kwargs: PMemo.kwargs,
    ) -> TMemo:
        """
        Memoize a function by storing its result in durable storage. This is useful for caching the results of expensive computations that you don't want to repeat on every workflow replay without needing to spawn a child workflow or set up an external cache. The function signature is intended to behave similarly to `asyncio.to_thread` or other similar uses of partially applied functions, where you pass in the function and its arguments separately.

        Note that memoization is performed at the _task run_ level, meaning you cannot cache across tasks (whether they're part of the same workflow or otherwise).

        :param fn: The function to compute the value to be memoized. This should be an async function that returns the value to be memoized.
        :param result_validator: The type of the result to be memoized. This is used for validating the result when it's retrieved from durable storage and for properly serializing the result of the function call. This is required and generally we recommend using either a Pydantic model, a dataclass, or a TypedDict, but you can also use `dict` as an escape hatch.
        :param *args: The arguments to pass to the function when computing the value to be memoized. These are used for computing the memoization key, so that different arguments will result in different cached values.
        :param **kwargs: The keyword arguments to pass to the function when computing the value to be memoized. These are used for computing the memoization key, so that different keyword arguments will result in different cached values.

        :return: The memoized value, either retrieved from durable storage or computed by calling the function.

        :raises TypeError: If the durable event listener is not of type DurableEventListener or PreEvictionDurableEventListener.
        """
        if not self._supports_durable_eviction:
            logger.warning(
                "Engine does not support memoization (requires >= %s). "
                "aio_memo will execute the function but results will not be "
                "persisted across replays. Upgrade your engine to enable durable memoization.",
                MinEngineVersion.DURABLE_EVICTION,
            )
            return await fn(*args, **kwargs)

        listener = self._durable_listener

        run_external_id = self.step_run_id
        adapter = TypeAdapter(result_validator)

        key = _compute_memo_key(self.step_run_id, *args, **kwargs)

        ack = await listener.send_event(
            durable_task_external_id=run_external_id,
            invocation_count=self.invocation_count,
            event=MemoEvent(memo_key=key, result=None),
        )

        if not isinstance(ack, DurableTaskEventMemoAck):
            raise TypeError(f"Expected memo ack, got {type(ack).__name__}")

        if ack.memo_already_existed and ack.memo_result_payload is None:
            logger.warning(
                "memo key found in durable storage but no data was returned. rerunning the function to recompute the value. "
            )

        if ack.memo_already_existed and ack.memo_result_payload is not None:
            serialized_result = ack.memo_result_payload
            result = adapter.validate_json(
                serialized_result, context=HATCHET_PYDANTIC_SENTINEL
            )
        else:
            result = await fn(*args, **kwargs)
            serialized_result = adapter.dump_json(
                result, context=HATCHET_PYDANTIC_SENTINEL
            )

            await self._ensure_stream_started()

            await listener.send_memo_completed_notification(
                durable_task_external_id=run_external_id,
                node_id=ack.node_id,
                branch_id=ack.branch_id,
                invocation_count=self.invocation_count,
                memo_result_payload=serialized_result,
                memo_key=key,
            )

        return result

    async def _now(self) -> MemoNowResult:
        ts = await asyncio.to_thread(datetime.now, timezone.utc)
        return MemoNowResult(ts=ts)

    async def aio_now(self) -> datetime:
        """
        Get the current timestamp. This is a wrapper around `datetime.now()` that is memoized using durable storage, so that it will return the same timestamp across replays of the same task run.

        :return: The current timestamp, memoized across replays of the same task run.
        """
        now = await self._aio_memo(
            self._now,
            MemoNowResult,
        )

        return now.ts

    async def aio_sleep_until(self, wake_at: datetime) -> SleepResult:
        """
        Durably sleep until a specific timestamp.

        :param wake_at: The timestamp to sleep until.

        :return: A SleepResult containing the actual duration slept, which may be different from the intended duration if the workflow was evicted and resumed.
        """
        now = await self.aio_now()

        return await self.aio_sleep_for(wake_at - now)



from hatchet_sdk.worker.durable_eviction.cache import DurableEvictionCache
from hatchet_sdk.worker.durable_eviction.manager import (
    DEFAULT_DURABLE_EVICTION_CONFIG,
    DurableEvictionConfig,
    DurableEvictionManager,
)

__all__ = [
    "DEFAULT_DURABLE_EVICTION_CONFIG",
    "DurableEvictionCache",
    "DurableEvictionConfig",
    "DurableEvictionManager",
]



import asyncio
from collections import OrderedDict
from collections.abc import Iterator
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from typing import Generic, TypeVar

K = TypeVar("K")
V = TypeVar("V")


class BoundedDict(OrderedDict[K, V]):
    def __init__(self, maxsize: int):
        super().__init__()
        self.maxsize = maxsize

    def __setitem__(self, key: K, value: V) -> None:
        if key in self:
            self.move_to_end(key)

        super().__setitem__(key, value)

        if len(self) > self.maxsize:
            self.popitem(last=False)


@dataclass
class TTLCacheEntry(Generic[V]):
    value: V
    expires_at: datetime


class TTLCache(Generic[K, V]):
    def __init__(self, ttl: timedelta) -> None:
        self.ttl = ttl
        self.cache: dict[K, TTLCacheEntry[V]] = {}

        self.eviction_job = asyncio.create_task(self._start_eviction_job())

    def __setitem__(self, key: K, value: V) -> None:
        self.cache[key] = TTLCacheEntry(
            value=value, expires_at=datetime.now(tz=timezone.utc) + self.ttl
        )

    def __getitem__(self, key: K) -> V:
        return self.cache[key].value

    def __contains__(self, key: object) -> bool:
        return key in self.cache

    def __delitem__(self, key: K) -> None:
        del self.cache[key]

    def __iter__(self) -> Iterator[K]:
        return iter(self.cache)

    def pop(self, key: K) -> V:
        return self.cache.pop(key).value

    def clear(self) -> None:
        self.cache.clear()

    def stop_eviction_job(self) -> None:
        self.eviction_job.cancel()

    async def _start_eviction_job(self) -> None:
        while True:
            await asyncio.sleep(self.ttl.total_seconds())

            now = datetime.now(tz=timezone.utc)
            expired = [k for k, entry in self.cache.items() if entry.expires_at <= now]

            for key in expired:
                del self.cache[key]
